"""
Smart TTL (Time-To-Live) management for form sessions.

TTL = clamp(minDays, effortDays, maxDays)
where effortDays = minutesSpent * effortMultiplier

Defaults: minDays=14, maxDays=90, effortMultiplier=0.5
"""

from __future__ import annotations

import time

from .types import (
    FORM_DEFINITION_DEFAULTS,
    FormDefinition,
    FormSession,
)

# Convenience accessors for nested defaults
_TTL_DEFAULTS: dict[str, float | int] = FORM_DEFINITION_DEFAULTS["ttl"]  # type: ignore[assignment]
_NUDGE_DEFAULTS: dict[str, float | int | bool] = FORM_DEFINITION_DEFAULTS["nudge"]  # type: ignore[assignment]


def _now_ms() -> int:
    """Current time in milliseconds since epoch."""
    return int(time.time() * 1000)


def calculate_ttl(session: FormSession, form: FormDefinition | None = None) -> int:
    """Calculate TTL based on user effort.

    Returns expiration timestamp in milliseconds since epoch.
    """
    ttl_cfg = form.ttl if form and form.ttl else None

    min_days: float = (ttl_cfg.min_days if ttl_cfg and ttl_cfg.min_days is not None else _TTL_DEFAULTS["min_days"])  # type: ignore[arg-type]
    max_days: float = (ttl_cfg.max_days if ttl_cfg and ttl_cfg.max_days is not None else _TTL_DEFAULTS["max_days"])  # type: ignore[arg-type]
    multiplier: float = (ttl_cfg.effort_multiplier if ttl_cfg and ttl_cfg.effort_multiplier is not None else _TTL_DEFAULTS["effort_multiplier"])  # type: ignore[arg-type]

    minutes_spent = session.effort.time_spent_ms / 60_000
    effort_days = minutes_spent * multiplier
    ttl_days = min(max_days, max(min_days, effort_days))

    return _now_ms() + int(ttl_days * 24 * 60 * 60 * 1000)


def should_nudge(session: FormSession, form: FormDefinition | None = None) -> bool:
    """Check if session should receive a nudge reminder."""
    nudge_cfg = form.nudge if form and form.nudge else None

    # Disabled?
    if nudge_cfg and nudge_cfg.enabled is False:
        return False

    # Max nudges reached?
    max_nudges: int = (nudge_cfg.max_nudges if nudge_cfg and nudge_cfg.max_nudges is not None else _NUDGE_DEFAULTS["max_nudges"])  # type: ignore[arg-type]
    if (session.nudge_count or 0) >= max_nudges:
        return False

    # Enough inactive time?
    after_hours: float = (nudge_cfg.after_inactive_hours if nudge_cfg and nudge_cfg.after_inactive_hours is not None else _NUDGE_DEFAULTS["after_inactive_hours"])  # type: ignore[arg-type]
    inactive_ms = after_hours * 60 * 60 * 1000
    now = _now_ms()
    if now - session.effort.last_interaction_at < inactive_ms:
        return False

    # Recent nudge? (24h minimum between nudges)
    if session.last_nudge_at is not None:
        if now - session.last_nudge_at < 24 * 60 * 60 * 1000:
            return False

    return True


def is_expiring_soon(session: FormSession, within_ms: int) -> bool:
    """Check if session expires within *within_ms* milliseconds."""
    return session.expires_at - _now_ms() < within_ms


def is_expired(session: FormSession) -> bool:
    """Check if session has expired."""
    return session.expires_at < _now_ms()


def should_confirm_cancel(session: FormSession) -> bool:
    """Check if we should confirm before cancelling (>5 min effort)."""
    return session.effort.time_spent_ms > 5 * 60 * 1000


def format_time_remaining(session: FormSession) -> str:
    """Human-readable time remaining for a session."""
    remaining = session.expires_at - _now_ms()

    if remaining <= 0:
        return "expired"

    hours = remaining // (60 * 60 * 1000)
    days = hours // 24

    if days > 0:
        return f"{days} day{'s' if days > 1 else ''}"
    if hours > 0:
        return f"{hours} hour{'s' if hours > 1 else ''}"

    minutes = remaining // (60 * 1000)
    return f"{minutes} minute{'s' if minutes > 1 else ''}"


def format_effort(session: FormSession) -> str:
    """Human-readable effort description."""
    minutes = session.effort.time_spent_ms // 60_000

    if minutes < 1:
        return "just started"
    if minutes < 60:
        return f"{minutes} minute{'s' if minutes > 1 else ''}"

    hours = minutes // 60
    remaining_minutes = minutes % 60

    if remaining_minutes == 0:
        return f"{hours} hour{'s' if hours > 1 else ''}"

    return f"{hours}h {remaining_minutes}m"



"""Tests for the builder module."""

from __future__ import annotations

import pytest

from elizaos_plugin_form.builder import C, ControlBuilder, Form, FormBuilder
from elizaos_plugin_form.types import FormControlOption


# ============================================================================
# CONTROL BUILDER – STATIC FACTORIES
# ============================================================================


class TestControlBuilderFactories:
    def test_text(self):
        ctrl = C.text("name").build()
        assert ctrl.key == "name"
        assert ctrl.type == "text"

    def test_email(self):
        ctrl = C.email("email").build()
        assert ctrl.type == "email"

    def test_number(self):
        ctrl = C.number("age").build()
        assert ctrl.type == "number"

    def test_boolean(self):
        ctrl = C.boolean_("agree").build()
        assert ctrl.type == "boolean"

    def test_select(self):
        opts = [FormControlOption(value="a", label="A")]
        ctrl = C.select("choice", opts).build()
        assert ctrl.type == "select"
        assert ctrl.options is not None
        assert len(ctrl.options) == 1

    def test_date(self):
        ctrl = C.date("dob").build()
        assert ctrl.type == "date"

    def test_file(self):
        ctrl = C.file("attachment").build()
        assert ctrl.type == "file"

    def test_field_generic(self):
        ctrl = C.field("custom").type("phone").build()
        assert ctrl.type == "phone"


# ============================================================================
# CONTROL BUILDER – CHAINING
# ============================================================================


class TestControlBuilderChaining:
    def test_required(self):
        ctrl = C.text("name").required().build()
        assert ctrl.required is True

    def test_optional(self):
        ctrl = C.text("name").required().optional().build()
        assert ctrl.required is False

    def test_hidden(self):
        ctrl = C.text("secret").hidden().build()
        assert ctrl.hidden is True

    def test_sensitive(self):
        ctrl = C.text("pwd").sensitive().build()
        assert ctrl.sensitive is True

    def test_readonly(self):
        ctrl = C.text("id").readonly().build()
        assert ctrl.readonly is True

    def test_multiple(self):
        ctrl = C.text("tags").multiple().build()
        assert ctrl.multiple is True

    def test_validation_chain(self):
        ctrl = (
            C.text("username")
            .required()
            .pattern(r"^[a-z]+$")
            .min_length(3)
            .max_length(20)
            .build()
        )
        assert ctrl.required is True
        assert ctrl.pattern == r"^[a-z]+$"
        assert ctrl.min_length == 3
        assert ctrl.max_length == 20

    def test_enum(self):
        ctrl = C.text("size").enum(["s", "m", "l"]).build()
        assert ctrl.enum == ["s", "m", "l"]

    def test_min_max(self):
        ctrl = C.number("age").min(0).max(150).build()
        assert ctrl.min == 0
        assert ctrl.max == 150

    def test_label(self):
        ctrl = C.text("fn").label("First Name").build()
        assert ctrl.label == "First Name"

    def test_ask(self):
        ctrl = C.text("email").ask("What is your email?").build()
        assert ctrl.ask_prompt == "What is your email?"

    def test_description(self):
        ctrl = C.text("bio").description("Tell us about yourself").build()
        assert ctrl.description == "Tell us about yourself"

    def test_hint(self):
        ctrl = C.text("wallet").hint("solana", "base58").build()
        assert ctrl.extract_hints == ["solana", "base58"]

    def test_example(self):
        ctrl = C.email("email").example("user@example.com").build()
        assert ctrl.example == "user@example.com"

    def test_confirm_threshold(self):
        ctrl = C.text("amount").confirm_threshold(0.95).build()
        assert ctrl.confirm_threshold == 0.95

    def test_file_options(self):
        ctrl = (
            C.file("docs")
            .accept(["application/pdf"])
            .max_size(1024)
            .max_files(3)
            .build()
        )
        assert ctrl.file is not None
        assert ctrl.file.accept == ["application/pdf"]
        assert ctrl.file.max_size == 1024
        assert ctrl.file.max_files == 3

    def test_roles(self):
        ctrl = C.text("discount").roles("admin", "sales").build()
        assert ctrl.roles == ["admin", "sales"]

    def test_default(self):
        ctrl = C.text("country").default("US").build()
        assert ctrl.default_value == "US"

    def test_depends_on(self):
        ctrl = C.text("state").depends_on("country", "equals", "US").build()
        assert ctrl.depends_on is not None
        assert ctrl.depends_on.field == "country"
        assert ctrl.depends_on.condition == "equals"
        assert ctrl.depends_on.value == "US"

    def test_dbbind(self):
        ctrl = C.text("email").dbbind("email_address").build()
        assert ctrl.dbbind == "email_address"

    def test_ui_options(self):
        ctrl = (
            C.text("name")
            .section("Personal")
            .order(1)
            .placeholder("Enter name")
            .help_text("Your full name")
            .widget("input")
            .build()
        )
        assert ctrl.ui is not None
        assert ctrl.ui.section == "Personal"
        assert ctrl.ui.order == 1
        assert ctrl.ui.placeholder == "Enter name"
        assert ctrl.ui.help_text == "Your full name"
        assert ctrl.ui.widget == "input"

    def test_i18n(self):
        ctrl = C.text("name").i18n("es", {"label": "Nombre"}).build()
        assert ctrl.i18n is not None
        assert "es" in ctrl.i18n
        assert ctrl.i18n["es"].label == "Nombre"

    def test_meta(self):
        ctrl = C.text("field").meta("priority", 1).build()
        assert ctrl.meta is not None
        assert ctrl.meta["priority"] == 1

    def test_auto_label_from_key(self):
        ctrl = C.text("first_name").build()
        assert ctrl.label == "First Name"


# ============================================================================
# FORM BUILDER
# ============================================================================


class TestFormBuilder:
    def test_create(self):
        form = Form.create("contact").build()
        assert form.id == "contact"
        assert form.name == "Contact"

    def test_name(self):
        form = Form.create("f").name("My Form").build()
        assert form.name == "My Form"

    def test_description(self):
        form = Form.create("f").description("A form").build()
        assert form.description == "A form"

    def test_version(self):
        form = Form.create("f").version(2).build()
        assert form.version == 2

    def test_control(self):
        form = Form.create("f").control(C.text("name")).build()
        assert len(form.controls) == 1
        assert form.controls[0].key == "name"

    def test_controls(self):
        form = (
            Form.create("f")
            .controls(C.text("name"), C.email("email"))
            .build()
        )
        assert len(form.controls) == 2

    def test_required_shorthand(self):
        form = Form.create("f").required("name", "email").build()
        assert len(form.controls) == 2
        assert all(c.required for c in form.controls)

    def test_optional_shorthand(self):
        form = Form.create("f").optional("notes").build()
        assert len(form.controls) == 1
        assert form.controls[0].required is False

    def test_roles(self):
        form = Form.create("f").roles("admin").build()
        assert form.roles == ["admin"]

    def test_allow_multiple(self):
        form = Form.create("f").allow_multiple().build()
        assert form.allow_multiple is True

    def test_no_undo(self):
        form = Form.create("f").no_undo().build()
        assert form.ux is not None
        assert form.ux.allow_undo is False

    def test_no_skip(self):
        form = Form.create("f").no_skip().build()
        assert form.ux is not None
        assert form.ux.allow_skip is False

    def test_no_autofill(self):
        form = Form.create("f").no_autofill().build()
        assert form.ux is not None
        assert form.ux.allow_autofill is False

    def test_max_undo_steps(self):
        form = Form.create("f").max_undo_steps(10).build()
        assert form.ux is not None
        assert form.ux.max_undo_steps == 10

    def test_ttl(self):
        form = Form.create("f").ttl(min_days=7, max_days=30).build()
        assert form.ttl is not None
        assert form.ttl.min_days == 7
        assert form.ttl.max_days == 30

    def test_no_nudge(self):
        form = Form.create("f").no_nudge().build()
        assert form.nudge is not None
        assert form.nudge.enabled is False

    def test_nudge_after(self):
        form = Form.create("f").nudge_after(24).build()
        assert form.nudge is not None
        assert form.nudge.after_inactive_hours == 24

    def test_nudge_message(self):
        form = Form.create("f").nudge_message("Hey!").build()
        assert form.nudge is not None
        assert form.nudge.message == "Hey!"

    def test_hooks(self):
        form = (
            Form.create("f")
            .on_start("start_worker")
            .on_field_change("field_worker")
            .on_ready("ready_worker")
            .on_submit("submit_worker")
            .on_cancel("cancel_worker")
            .on_expire("expire_worker")
            .build()
        )
        assert form.hooks is not None
        assert form.hooks.on_start == "start_worker"
        assert form.hooks.on_field_change == "field_worker"
        assert form.hooks.on_ready == "ready_worker"
        assert form.hooks.on_submit == "submit_worker"
        assert form.hooks.on_cancel == "cancel_worker"
        assert form.hooks.on_expire == "expire_worker"

    def test_debug(self):
        form = Form.create("f").debug().build()
        assert form.debug is True

    def test_i18n(self):
        form = Form.create("f").i18n("es", {"name": "Formulario"}).build()
        assert form.i18n is not None
        assert "es" in form.i18n
        assert form.i18n["es"].name == "Formulario"

    def test_meta(self):
        form = Form.create("f").meta("category", "support").build()
        assert form.meta is not None
        assert form.meta["category"] == "support"


# ============================================================================
# ALIASES
# ============================================================================


class TestAliases:
    def test_form_alias(self):
        assert Form is FormBuilder

    def test_c_alias(self):
        assert C is ControlBuilder



"""
Field validation utilities for the Form Plugin.

Validation happens at two points:
1. At Extraction – immediate feedback on extracted values
2. At Submission – safety net before final submission

Custom type handlers are checked FIRST, before built-in type validation.
"""

from __future__ import annotations

import math
import re
from datetime import datetime
from typing import Callable, Protocol

from .types import (
    FormControl,
    FormControlFileOptions,
    JsonValue,
    ValidationResult,
)


# ============================================================================
# TYPE HANDLER PROTOCOL
# ============================================================================


class TypeHandlerProtocol(Protocol):
    """Protocol for custom type handlers."""

    def validate(self, value: JsonValue, control: FormControl) -> ValidationResult: ...
    def parse(self, value: str) -> JsonValue: ...
    def format(self, value: JsonValue) -> str: ...


class TypeHandler:
    """Custom type handler with optional validate/parse/format/extraction_prompt."""

    def __init__(
        self,
        validate: Callable[[JsonValue, FormControl], ValidationResult] | None = None,
        parse: Callable[[str], JsonValue] | None = None,
        format: Callable[[JsonValue], str] | None = None,
        extraction_prompt: str | None = None,
    ) -> None:
        self._validate = validate
        self._parse = parse
        self._format = format
        self.extraction_prompt = extraction_prompt

    def validate(self, value: JsonValue, control: FormControl) -> ValidationResult:
        if self._validate is not None:
            return self._validate(value, control)
        return ValidationResult(valid=True)

    def parse(self, value: str) -> JsonValue:
        if self._parse is not None:
            return self._parse(value)
        return value

    def format(self, value: JsonValue) -> str:
        if self._format is not None:
            return self._format(value)
        return str(value)


# ============================================================================
# TYPE HANDLER REGISTRY
# ============================================================================

_type_handlers: dict[str, TypeHandler] = {}


def register_type_handler(type_name: str, handler: TypeHandler) -> None:
    """Register a custom type handler."""
    _type_handlers[type_name] = handler


def get_type_handler(type_name: str) -> TypeHandler | None:
    """Get a type handler or ``None`` if not registered."""
    return _type_handlers.get(type_name)


def clear_type_handlers() -> None:
    """Clear all type handlers (for tests)."""
    _type_handlers.clear()


# ============================================================================
# FIELD VALIDATION
# ============================================================================


def validate_field(value: JsonValue, control: FormControl) -> ValidationResult:
    """Validate a value against a control's validation rules.

    Order: required → custom handler → type-specific → pattern/limits.
    """
    # Required check
    if control.required:
        if value is None or value == "":
            label = control.label or control.key
            return ValidationResult(valid=False, error=f"{label} is required")

    # Empty optional fields are valid
    if value is None or value == "":
        return ValidationResult(valid=True)

    # Custom type handler first
    handler = _type_handlers.get(control.type)
    if handler is not None and handler._validate is not None:
        result = handler.validate(value, control)
        if not result.valid:
            return result

    # Type-specific validation
    type_map: dict[str, Callable[[JsonValue, FormControl], ValidationResult]] = {
        "email": _validate_email,
        "number": _validate_number,
        "boolean": _validate_boolean,
        "date": _validate_date,
        "select": _validate_select,
        "file": _validate_file,
    }

    validator = type_map.get(control.type, _validate_text)
    return validator(value, control)


# ============================================================================
# TYPE-SPECIFIC VALIDATORS
# ============================================================================

_EMAIL_REGEX = re.compile(r"^[^\s@]+@[^\s@]+\.[^\s@]+$")


def _validate_email(value: JsonValue, control: FormControl) -> ValidationResult:
    str_value = str(value)
    if not _EMAIL_REGEX.match(str_value):
        label = control.label or control.key
        return ValidationResult(valid=False, error=f"{label} must be a valid email address")
    return _validate_text(value, control)


def _validate_number(value: JsonValue, control: FormControl) -> ValidationResult:
    if isinstance(value, (int, float)) and not isinstance(value, bool):
        num_value = float(value)
    else:
        cleaned = re.sub(r"[,$]", "", str(value))
        try:
            num_value = float(cleaned)
        except (ValueError, TypeError):
            label = control.label or control.key
            return ValidationResult(valid=False, error=f"{label} must be a number")

    if math.isnan(num_value):
        label = control.label or control.key
        return ValidationResult(valid=False, error=f"{label} must be a number")

    if control.min is not None and num_value < control.min:
        label = control.label or control.key
        return ValidationResult(valid=False, error=f"{label} must be at least {control.min}")

    if control.max is not None and num_value > control.max:
        label = control.label or control.key
        return ValidationResult(valid=False, error=f"{label} must be at most {control.max}")

    return ValidationResult(valid=True)


def _validate_boolean(value: JsonValue, control: FormControl) -> ValidationResult:
    if isinstance(value, bool):
        return ValidationResult(valid=True)

    str_value = str(value).lower()
    truthy = {"true", "yes", "1", "on"}
    falsy = {"false", "no", "0", "off"}

    if str_value in truthy or str_value in falsy:
        return ValidationResult(valid=True)

    return ValidationResult(valid=False, error="Must be true or false")


def _validate_date(value: JsonValue, control: FormControl) -> ValidationResult:
    date_value: datetime | None = None
    timestamp: float | None = None

    if isinstance(value, (int, float)) and not isinstance(value, bool):
        try:
            date_value = datetime.fromtimestamp(value / 1000)
            timestamp = float(value)
        except (OSError, ValueError, OverflowError):
            label = control.label or control.key
            return ValidationResult(valid=False, error=f"{label} must be a valid date")
    elif isinstance(value, str):
        # Try ISO parse
        for fmt in ("%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d"):
            try:
                date_value = datetime.strptime(value, fmt)
                timestamp = date_value.timestamp() * 1000
                break
            except ValueError:
                continue
        if date_value is None:
            label = control.label or control.key
            return ValidationResult(valid=False, error=f"{label} must be a valid date")
    else:
        label = control.label or control.key
        return ValidationResult(valid=False, error=f"{label} must be a valid date")

    if timestamp is not None:
        if control.min is not None and timestamp < control.min:
            label = control.label or control.key
            return ValidationResult(valid=False, error=f"{label} is too early")
        if control.max is not None and timestamp > control.max:
            label = control.label or control.key
            return ValidationResult(valid=False, error=f"{label} is too late")

    return ValidationResult(valid=True)


def _validate_select(value: JsonValue, control: FormControl) -> ValidationResult:
    if not control.options or len(control.options) == 0:
        return ValidationResult(valid=True)

    str_value = str(value)
    valid_values = [opt.value for opt in control.options]

    if str_value not in valid_values:
        label = control.label or control.key
        return ValidationResult(
            valid=False,
            error=f"{label} must be one of the available options",
        )

    return ValidationResult(valid=True)


def _validate_text(value: JsonValue, control: FormControl) -> ValidationResult:
    str_value = str(value)
    label = control.label or control.key

    # Pattern
    if control.pattern:
        if not re.search(control.pattern, str_value):
            return ValidationResult(valid=False, error=f"{label} has invalid format")

    # Length
    if control.min_length is not None and len(str_value) < control.min_length:
        return ValidationResult(
            valid=False,
            error=f"{label} must be at least {control.min_length} characters",
        )

    if control.max_length is not None and len(str_value) > control.max_length:
        return ValidationResult(
            valid=False,
            error=f"{label} must be at most {control.max_length} characters",
        )

    # Enum
    if control.enum and len(control.enum) > 0:
        if str_value not in control.enum:
            return ValidationResult(
                valid=False,
                error=f"{label} must be one of: {', '.join(control.enum)}",
            )

    return ValidationResult(valid=True)


def _validate_file(value: JsonValue, control: FormControl) -> ValidationResult:
    if control.file is None:
        return ValidationResult(valid=True)

    files = value if isinstance(value, list) else [value]

    # Max files
    if control.file.max_files is not None and len(files) > control.file.max_files:
        return ValidationResult(
            valid=False,
            error=f"Maximum {control.file.max_files} files allowed",
        )

    for f in files:
        if not f or not isinstance(f, dict):
            continue

        # Max size
        if control.file.max_size is not None:
            size = f.get("size")
            if isinstance(size, (int, float)) and size > control.file.max_size:
                return ValidationResult(
                    valid=False,
                    error=f"File size exceeds maximum of {_format_bytes(control.file.max_size)}",
                )

        # MIME types
        if control.file.accept:
            mime_type = f.get("mimeType") or f.get("mime_type")
            if mime_type:
                accepted = any(matches_mime_type(mime_type, p) for p in control.file.accept)
                if not accepted:
                    return ValidationResult(
                        valid=False,
                        error=f"File type {mime_type} is not accepted",
                    )

    return ValidationResult(valid=True)


# ============================================================================
# HELPERS
# ============================================================================


def matches_mime_type(mime_type: str, pattern: str) -> bool:
    """Check if a MIME type matches a pattern.

    Supports exact match (``image/png``), wildcard (``image/*``),
    and universal (``*/*``).
    """
    if pattern == "*/*":
        return True
    if pattern.endswith("/*"):
        prefix = pattern[:-1]  # "image/" from "image/*"
        return mime_type.startswith(prefix)
    return mime_type == pattern


def _format_bytes(num_bytes: int) -> str:
    if num_bytes < 1024:
        return f"{num_bytes} B"
    if num_bytes < 1024 * 1024:
        return f"{num_bytes / 1024:.1f} KB"
    if num_bytes < 1024 * 1024 * 1024:
        return f"{num_bytes / (1024 * 1024):.1f} MB"
    return f"{num_bytes / (1024 * 1024 * 1024):.1f} GB"


# ============================================================================
# VALUE PARSING
# ============================================================================


def parse_value(value: str, control: FormControl) -> JsonValue:
    """Parse a string value to the appropriate type based on control type."""
    handler = _type_handlers.get(control.type)
    if handler is not None and handler._parse is not None:
        return handler.parse(value)

    if control.type == "number":
        cleaned = re.sub(r"[,$]", "", value)
        try:
            return float(cleaned)
        except ValueError:
            return value

    if control.type == "boolean":
        return value.lower() in {"true", "yes", "1", "on"}

    if control.type == "date":
        for fmt in ("%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d"):
            try:
                dt = datetime.strptime(value, fmt)
                return dt.strftime("%Y-%m-%dT%H:%M:%S") + "Z" if "T" in value else dt.strftime("%Y-%m-%d")
            except ValueError:
                continue
        return value

    # text, email, select, default
    return value


# ============================================================================
# VALUE FORMATTING
# ============================================================================


def format_value(value: JsonValue, control: FormControl) -> str:
    """Format a value for display."""
    if value is None:
        return ""

    handler = _type_handlers.get(control.type)
    if handler is not None and handler._format is not None:
        return handler.format(value)

    # Sensitive masking
    if control.sensitive:
        str_val = str(value)
        if len(str_val) > 8:
            return f"{str_val[:4]}...{str_val[-4:]}"
        return "****"

    if control.type == "number":
        if isinstance(value, (int, float)) and not isinstance(value, bool):
            return f"{value:,}" if isinstance(value, int) else f"{value:,.2f}" if value != int(value) else f"{int(value):,}"
        return str(value)

    if control.type == "boolean":
        return "Yes" if value else "No"

    if control.type == "date":
        return str(value)

    if control.type == "select":
        if control.options:
            for opt in control.options:
                if opt.value == str(value):
                    return opt.label
        return str(value)

    if control.type == "file":
        if isinstance(value, list):
            return ", ".join(
                (f.get("name", "file") if isinstance(f, dict) else "file") for f in value
            )
        if isinstance(value, dict):
            return value.get("name", "file")  # type: ignore[return-value]
        return "file"

    return str(value)



"""
Built-in control types for the Form Plugin.

Seven standard types: text, number, email, boolean, select, date, file.
"""

from __future__ import annotations

import re
from typing import Callable

from .types import FormControl, JsonValue, ValidationResult

# ---------------------------------------------------------------------------
# ControlType dict shape
# ---------------------------------------------------------------------------
ControlType = dict[str, object]

# ---------------------------------------------------------------------------
# Regex helpers
# ---------------------------------------------------------------------------
_EMAIL_REGEX = re.compile(r"^[^\s@]+@[^\s@]+\.[^\s@]+$")
_ISO_DATE_REGEX = re.compile(r"^\d{4}-\d{2}-\d{2}$")

# ============================================================================
# BUILT-IN TYPES
# ============================================================================

_text_type: ControlType = {
    "id": "text",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None else
        _text_validate(value, control)
    ),
    "parse": lambda value: str(value).strip(),
    "format": lambda value: str(value) if value is not None else "",
    "extraction_prompt": "a text string",
}


def _text_validate(value: JsonValue, control: FormControl) -> ValidationResult:
    s = str(value)
    if control.min_length is not None and len(s) < control.min_length:
        return ValidationResult(valid=False, error=f"Must be at least {control.min_length} characters")
    if control.max_length is not None and len(s) > control.max_length:
        return ValidationResult(valid=False, error=f"Must be at most {control.max_length} characters")
    if control.pattern and not re.search(control.pattern, s):
        return ValidationResult(valid=False, error="Invalid format")
    if control.enum and s not in control.enum:
        return ValidationResult(valid=False, error=f"Must be one of: {', '.join(control.enum)}")
    return ValidationResult(valid=True)


_number_type: ControlType = {
    "id": "number",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None or value == "" else
        _number_validate(value, control)
    ),
    "parse": lambda value: float(re.sub(r"[,$\s]", "", value)),
    "format": lambda value: (
        "" if value is None else
        f"{value:,}" if isinstance(value, int) else
        str(value)
    ),
    "extraction_prompt": "a number (integer or decimal)",
}


def _number_validate(value: JsonValue, control: FormControl) -> ValidationResult:
    if isinstance(value, (int, float)) and not isinstance(value, bool):
        num = float(value)
    else:
        try:
            num = float(str(value))
        except (ValueError, TypeError):
            return ValidationResult(valid=False, error="Must be a valid number")

    import math
    if math.isnan(num):
        return ValidationResult(valid=False, error="Must be a valid number")
    if control.min is not None and num < control.min:
        return ValidationResult(valid=False, error=f"Must be at least {control.min}")
    if control.max is not None and num > control.max:
        return ValidationResult(valid=False, error=f"Must be at most {control.max}")
    return ValidationResult(valid=True)


_email_type: ControlType = {
    "id": "email",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None or value == "" else
        (
            ValidationResult(valid=False, error="Invalid email format")
            if not _EMAIL_REGEX.match(str(value).strip().lower())
            else ValidationResult(valid=True)
        )
    ),
    "parse": lambda value: value.strip().lower(),
    "format": lambda value: str(value or "").lower(),
    "extraction_prompt": "an email address (e.g., user@example.com)",
}


_boolean_type: ControlType = {
    "id": "boolean",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None else
        (
            ValidationResult(valid=True) if isinstance(value, bool) else
            (
                ValidationResult(valid=True)
                if str(value).lower() in {"true", "false", "yes", "no", "1", "0", "on", "off"}
                else ValidationResult(valid=False, error="Must be yes/no or true/false")
            )
        )
    ),
    "parse": lambda value: str(value).lower() in {"true", "yes", "1", "on"},
    "format": lambda value: (
        "Yes" if value is True else
        "No" if value is False else
        str(value or "")
    ),
    "extraction_prompt": "a yes/no or true/false value",
}


_select_type: ControlType = {
    "id": "select",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None or value == "" else
        _select_validate(value, control)
    ),
    "parse": lambda value: value.strip(),
    "format": lambda value: str(value or ""),
    "extraction_prompt": "one of the available options",
}


def _select_validate(value: JsonValue, control: FormControl) -> ValidationResult:
    s = str(value)
    if control.options:
        valid_vals = [o.value for o in control.options]
        if s not in valid_vals:
            labels = ", ".join(o.label for o in control.options)
            return ValidationResult(valid=False, error=f"Must be one of: {labels}")
    if control.enum and not control.options:
        if s not in control.enum:
            return ValidationResult(valid=False, error=f"Must be one of: {', '.join(control.enum)}")
    return ValidationResult(valid=True)


_date_type: ControlType = {
    "id": "date",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None or value == "" else
        _date_validate(value, control)
    ),
    "parse": lambda value: _date_parse(value),
    "format": lambda value: str(value) if value else "",
    "extraction_prompt": "a date (preferably in YYYY-MM-DD format)",
}


def _date_validate(value: JsonValue, control: FormControl) -> ValidationResult:
    s = str(value)
    if not _ISO_DATE_REGEX.match(s):
        return ValidationResult(valid=False, error="Must be in YYYY-MM-DD format")
    from datetime import datetime
    try:
        datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        return ValidationResult(valid=False, error="Invalid date")
    return ValidationResult(valid=True)


def _date_parse(value: str) -> str:
    from datetime import datetime
    for fmt in ("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%SZ"):
        try:
            dt = datetime.strptime(value.strip(), fmt)
            return dt.strftime("%Y-%m-%d")
        except ValueError:
            continue
    return value.strip()


_file_type: ControlType = {
    "id": "file",
    "builtin": True,
    "validate": lambda value, control: (
        ValidationResult(valid=True) if value is None else
        (
            ValidationResult(valid=True) if isinstance(value, (dict, list)) else
            ValidationResult(valid=False, error="Invalid file data")
        )
    ),
    "format": lambda value: (
        "" if not value else
        f"{len(value)} file(s)" if isinstance(value, list) else
        (value.get("name", "File attached") if isinstance(value, dict) else "File attached")
    ),
    "extraction_prompt": "a file attachment (upload required)",
}


# ============================================================================
# EXPORTS
# ============================================================================

BUILTIN_TYPES: list[ControlType] = [
    _text_type,
    _number_type,
    _email_type,
    _boolean_type,
    _select_type,
    _date_type,
    _file_type,
]

BUILTIN_TYPE_MAP: dict[str, ControlType] = {str(t["id"]): t for t in BUILTIN_TYPES}


def register_builtin_types(register_fn: Callable[[ControlType], None]) -> None:
    """Register all built-in types with the given function."""
    for t in BUILTIN_TYPES:
        register_fn(t)


def get_builtin_type(type_id: str) -> ControlType | None:
    """Get a built-in type by id."""
    return BUILTIN_TYPE_MAP.get(type_id)


def is_builtin_type(type_id: str) -> bool:
    """Check if a type id is a built-in type."""
    return type_id in BUILTIN_TYPE_MAP



"""Tests for the defaults module."""

from __future__ import annotations

import pytest

from elizaos_plugin_form.defaults import apply_control_defaults, apply_form_defaults, prettify


# ============================================================================
# PRETTIFY
# ============================================================================


class TestPrettify:
    def test_snake_case(self):
        assert prettify("first_name") == "First Name"

    def test_kebab_case(self):
        assert prettify("email-address") == "Email Address"

    def test_single_word(self):
        assert prettify("email") == "Email"

    def test_multiple_underscores(self):
        assert prettify("home_phone_number") == "Home Phone Number"

    def test_already_title_case(self):
        assert prettify("Email") == "Email"

    def test_empty_string(self):
        assert prettify("") == ""

    def test_mixed_separators(self):
        assert prettify("user_first-name") == "User First Name"


# ============================================================================
# APPLY CONTROL DEFAULTS
# ============================================================================


class TestApplyControlDefaults:
    def test_minimal_control(self):
        result = apply_control_defaults({"key": "email"})
        assert result["key"] == "email"
        assert result["label"] == "Email"
        assert result["type"] == "text"
        assert result["required"] is False
        assert result["confirm_threshold"] == 0.8

    def test_custom_label_preserved(self):
        result = apply_control_defaults({"key": "email", "label": "Your Email"})
        assert result["label"] == "Your Email"

    def test_custom_type_preserved(self):
        result = apply_control_defaults({"key": "email", "type": "email"})
        assert result["type"] == "email"

    def test_required_preserved(self):
        result = apply_control_defaults({"key": "name", "required": True})
        assert result["required"] is True

    def test_extra_fields_preserved(self):
        result = apply_control_defaults({
            "key": "age",
            "type": "number",
            "min": 0,
            "max": 150,
            "description": "Your age",
        })
        assert result["min"] == 0
        assert result["max"] == 150
        assert result["description"] == "Your age"

    def test_confirm_threshold_override(self):
        result = apply_control_defaults({"key": "k", "confirm_threshold": 0.95})
        assert result["confirm_threshold"] == 0.95


# ============================================================================
# APPLY FORM DEFAULTS
# ============================================================================


class TestApplyFormDefaults:
    def test_minimal_form(self):
        result = apply_form_defaults({"id": "contact"})
        assert result["id"] == "contact"
        assert result["name"] == "Contact"
        assert result["version"] == 1
        assert result["status"] == "active"
        assert result["debug"] is False
        # UX defaults
        ux = result["ux"]
        assert ux["allow_undo"] is True
        assert ux["allow_skip"] is True
        assert ux["max_undo_steps"] == 5
        # TTL defaults
        ttl = result["ttl"]
        assert ttl["min_days"] == 14
        assert ttl["max_days"] == 90
        assert ttl["effort_multiplier"] == 0.5
        # Nudge defaults
        nudge = result["nudge"]
        assert nudge["enabled"] is True
        assert nudge["after_inactive_hours"] == 48
        assert nudge["max_nudges"] == 3

    def test_custom_name_preserved(self):
        result = apply_form_defaults({"id": "contact", "name": "Contact Us"})
        assert result["name"] == "Contact Us"

    def test_custom_version_preserved(self):
        result = apply_form_defaults({"id": "contact", "version": 5})
        assert result["version"] == 5

    def test_controls_get_defaults(self):
        result = apply_form_defaults({
            "id": "form1",
            "controls": [{"key": "email"}],
        })
        controls = result["controls"]
        assert len(controls) == 1
        assert controls[0]["label"] == "Email"
        assert controls[0]["type"] == "text"

    def test_partial_ux_merged(self):
        result = apply_form_defaults({
            "id": "form1",
            "ux": {"allow_undo": False},
        })
        ux = result["ux"]
        assert ux["allow_undo"] is False
        assert ux["allow_skip"] is True  # default

    def test_partial_ttl_merged(self):
        result = apply_form_defaults({
            "id": "form1",
            "ttl": {"min_days": 7},
        })
        ttl = result["ttl"]
        assert ttl["min_days"] == 7
        assert ttl["max_days"] == 90  # default

    def test_partial_nudge_merged(self):
        result = apply_form_defaults({
            "id": "form1",
            "nudge": {"max_nudges": 5},
        })
        nudge = result["nudge"]
        assert nudge["max_nudges"] == 5
        assert nudge["enabled"] is True  # default

    def test_extra_fields_preserved(self):
        result = apply_form_defaults({
            "id": "form1",
            "description": "A test form",
            "roles": ["admin"],
        })
        assert result["description"] == "A test form"
        assert result["roles"] == ["admin"]



"""
Default value application for forms and controls.

Ensures minimal definitions work correctly by filling in sensible defaults.
"""

from __future__ import annotations

import re
from copy import deepcopy
from typing import cast

from .types import (
    FORM_CONTROL_DEFAULTS,
    FORM_DEFINITION_DEFAULTS,
    FormControl,
    FormDefinition,
    FormDefinitionNudge,
    FormDefinitionTTL,
    FormDefinitionUX,
    JsonValue,
)


def prettify(key: str) -> str:
    """Convert snake_case or kebab-case to Title Case.

    >>> prettify("first_name")
    'First Name'
    >>> prettify("email-address")
    'Email Address'
    """
    spaced = re.sub(r"[-_]", " ", key)
    return re.sub(r"\b\w", lambda m: m.group().upper(), spaced)


def apply_control_defaults(control: dict[str, JsonValue] | FormControl) -> dict[str, JsonValue]:
    """Apply defaults to a control dict and return a complete dict."""
    if isinstance(control, FormControl):
        src: dict[str, JsonValue] = {
            "key": control.key,
            "label": control.label,
            "type": control.type,
            "required": control.required,
        }
        if control.confirm_threshold is not None:
            src["confirm_threshold"] = control.confirm_threshold
        if control.pattern is not None:
            src["pattern"] = control.pattern
        if control.min is not None:
            src["min"] = control.min
        if control.max is not None:
            src["max"] = control.max
        if control.description is not None:
            src["description"] = control.description
    else:
        src = dict(control)

    key = str(src.get("key", ""))
    result: dict[str, JsonValue] = {
        "key": key,
        "label": src.get("label") or prettify(key),
        "type": src.get("type") or cast(str, FORM_CONTROL_DEFAULTS["type"]),
        "required": src.get("required") if src.get("required") is not None else FORM_CONTROL_DEFAULTS["required"],
        "confirm_threshold": (
            src.get("confirm_threshold")
            if src.get("confirm_threshold") is not None
            else FORM_CONTROL_DEFAULTS["confirm_threshold"]
        ),
    }

    # Merge remaining keys from source
    for k, v in src.items():
        if k not in result:
            result[k] = v

    return result


def apply_form_defaults(form: dict[str, JsonValue] | FormDefinition) -> dict[str, JsonValue]:
    """Apply defaults to a form dict and return a complete dict."""
    _def = FORM_DEFINITION_DEFAULTS
    _ux_def: dict[str, JsonValue] = _def["ux"]  # type: ignore[assignment]
    _ttl_def: dict[str, JsonValue] = _def["ttl"]  # type: ignore[assignment]
    _nudge_def: dict[str, JsonValue] = _def["nudge"]  # type: ignore[assignment]

    if isinstance(form, FormDefinition):
        src: dict[str, JsonValue] = {"id": form.id, "name": form.name}
        if form.version is not None:
            src["version"] = form.version
        if form.status is not None:
            src["status"] = form.status
        if form.controls is not None:
            src["controls"] = form.controls  # type: ignore[assignment]
        if form.debug is not None:
            src["debug"] = form.debug
        if form.description is not None:
            src["description"] = form.description
    else:
        src = dict(form)

    form_id = str(src.get("id", ""))

    # UX
    ux_src = src.get("ux") or {}
    if isinstance(ux_src, dict):
        ux = {
            "allow_undo": ux_src.get("allow_undo") if ux_src.get("allow_undo") is not None else _ux_def["allow_undo"],
            "allow_skip": ux_src.get("allow_skip") if ux_src.get("allow_skip") is not None else _ux_def["allow_skip"],
            "max_undo_steps": ux_src.get("max_undo_steps") if ux_src.get("max_undo_steps") is not None else _ux_def["max_undo_steps"],
            "show_examples": ux_src.get("show_examples") if ux_src.get("show_examples") is not None else _ux_def["show_examples"],
            "show_explanations": ux_src.get("show_explanations") if ux_src.get("show_explanations") is not None else _ux_def["show_explanations"],
            "allow_autofill": ux_src.get("allow_autofill") if ux_src.get("allow_autofill") is not None else _ux_def["allow_autofill"],
        }
    else:
        ux = dict(_ux_def)

    # TTL
    ttl_src = src.get("ttl") or {}
    if isinstance(ttl_src, dict):
        ttl = {
            "min_days": ttl_src.get("min_days") if ttl_src.get("min_days") is not None else _ttl_def["min_days"],
            "max_days": ttl_src.get("max_days") if ttl_src.get("max_days") is not None else _ttl_def["max_days"],
            "effort_multiplier": ttl_src.get("effort_multiplier") if ttl_src.get("effort_multiplier") is not None else _ttl_def["effort_multiplier"],
        }
    else:
        ttl = dict(_ttl_def)

    # Nudge
    nudge_src = src.get("nudge") or {}
    if isinstance(nudge_src, dict):
        nudge: dict[str, JsonValue] = {
            "enabled": nudge_src.get("enabled") if nudge_src.get("enabled") is not None else _nudge_def["enabled"],
            "after_inactive_hours": nudge_src.get("after_inactive_hours") if nudge_src.get("after_inactive_hours") is not None else _nudge_def["after_inactive_hours"],
            "max_nudges": nudge_src.get("max_nudges") if nudge_src.get("max_nudges") is not None else _nudge_def["max_nudges"],
        }
        if nudge_src.get("message") is not None:
            nudge["message"] = nudge_src["message"]
    else:
        nudge = dict(_nudge_def)

    # Controls
    raw_controls = src.get("controls") or []
    controls: list[dict[str, JsonValue]] = []
    if isinstance(raw_controls, list):
        for c in raw_controls:
            if isinstance(c, (dict, FormControl)):
                controls.append(apply_control_defaults(c))
            # else skip non-dict items

    result: dict[str, JsonValue] = {
        "id": form_id,
        "name": src.get("name") or prettify(form_id),
        "version": src.get("version") if src.get("version") is not None else _def["version"],
        "status": src.get("status") if src.get("status") is not None else _def["status"],
        "controls": controls,  # type: ignore[dict-item]
        "ux": ux,  # type: ignore[dict-item]
        "ttl": ttl,  # type: ignore[dict-item]
        "nudge": nudge,  # type: ignore[dict-item]
        "debug": src.get("debug") if src.get("debug") is not None else _def["debug"],
    }

    # Merge remaining keys
    for k, v in src.items():
        if k not in result:
            result[k] = v

    return result



"""
Simple template resolution for form-controlled prompts.

Templates use ``{{ variable }}`` syntax (Mustache-like).
"""

from __future__ import annotations

import re
from copy import deepcopy

from .types import FormControl, FormSession

TEMPLATE_PATTERN = r"\{\{\s*([a-zA-Z0-9_.-]+)\s*\}\}"

_COMPILED_PATTERN = re.compile(TEMPLATE_PATTERN)


def build_template_values(session: FormSession) -> dict[str, str]:
    """Build a values dict from session fields and context."""
    values: dict[str, str] = {}

    for key, state in session.fields.items():
        val = state.value
        if isinstance(val, str):
            values[key] = val
        elif isinstance(val, (int, float, bool)):
            values[key] = str(val)

    if session.context and isinstance(session.context, dict):
        for key, val in session.context.items():
            if isinstance(val, str):
                values[key] = val
            elif isinstance(val, (int, float, bool)):
                values[key] = str(val)

    return values


def render_template(
    template: str | None,
    values: dict[str, str],
) -> str | None:
    """Render a template string, substituting ``{{ key }}`` placeholders.

    Returns ``None`` if *template* is ``None``.  Unresolved placeholders
    are left in place.
    """
    if template is None:
        return None

    def _replacer(m: re.Match[str]) -> str:
        key = m.group(1)
        replacement = values.get(key)
        return replacement if replacement is not None else m.group(0)

    return _COMPILED_PATTERN.sub(_replacer, template)


def resolve_control_templates(
    control: FormControl,
    values: dict[str, str],
) -> FormControl:
    """Return a *new* ``FormControl`` with templates resolved."""
    # Resolve simple string fields
    new_label = render_template(control.label, values) or control.label
    new_description = render_template(control.description, values)
    new_ask_prompt = render_template(control.ask_prompt, values)
    new_example = render_template(control.example, values)

    # Resolve extract hints
    new_hints: list[str] | None = None
    if control.extract_hints:
        new_hints = [render_template(h, values) or h for h in control.extract_hints]

    # Resolve options
    new_options = None
    if control.options:
        from .types import FormControlOption

        new_options = []
        for opt in control.options:
            new_options.append(
                FormControlOption(
                    value=opt.value,
                    label=render_template(opt.label, values) or opt.label,
                    description=render_template(opt.description, values),
                )
            )

    # Resolve nested fields
    new_fields = None
    if control.fields:
        new_fields = [resolve_control_templates(f, values) for f in control.fields]

    return FormControl(
        key=control.key,
        label=new_label,
        type=control.type,
        required=control.required,
        multiple=control.multiple,
        readonly=control.readonly,
        hidden=control.hidden,
        sensitive=control.sensitive,
        dbbind=control.dbbind,
        pattern=control.pattern,
        min=control.min,
        max=control.max,
        min_length=control.min_length,
        max_length=control.max_length,
        enum=control.enum,
        options=new_options if new_options is not None else control.options,
        file=control.file,
        default_value=control.default_value,
        depends_on=control.depends_on,
        roles=control.roles,
        description=new_description,
        ask_prompt=new_ask_prompt,
        extract_hints=new_hints if new_hints is not None else control.extract_hints,
        confirm_threshold=control.confirm_threshold,
        example=new_example,
        ui=control.ui,
        i18n=control.i18n,
        fields=new_fields if new_fields is not None else control.fields,
        meta=control.meta,
    )



"""Comprehensive tests for the intent module."""

from __future__ import annotations

import pytest

from elizaos_plugin_form.intent import (
    has_data_to_extract,
    is_lifecycle_intent,
    is_ux_intent,
    quick_intent_detect,
)


# ============================================================================
# LIFECYCLE INTENTS
# ============================================================================


class TestRestoreIntent:
    @pytest.mark.parametrize("text", [
        "resume",
        "continue",
        "pick up where I left off",
        "go back to my form",
        "get back to my form",
        "I want to resume",
        "Can I continue?",
    ])
    def test_restore_keywords(self, text):
        assert quick_intent_detect(text) == "restore"


class TestSubmitIntent:
    @pytest.mark.parametrize("text", [
        "submit",
        "done",
        "finish",
        "send it",
        "that's all",
        "thats all",
        "i'm done",
        "im done",
        "complete",
        "all set",
        "I'm done with the form",
    ])
    def test_submit_keywords(self, text):
        assert quick_intent_detect(text) == "submit"


class TestStashIntent:
    @pytest.mark.parametrize("text", [
        "save",
        "stash",
        "later",
        "hold on",
        "pause",
        "save for later",
        "come back",
        "save this",
    ])
    def test_stash_keywords(self, text):
        assert quick_intent_detect(text) == "stash"

    def test_save_and_submit_excluded(self):
        """'save and submit' should NOT trigger stash."""
        result = quick_intent_detect("save and submit")
        assert result != "stash"

    def test_save_and_send_excluded(self):
        result = quick_intent_detect("save and send")
        assert result != "stash"


class TestCancelIntent:
    @pytest.mark.parametrize("text", [
        "cancel",
        "abort",
        "nevermind",
        "never mind",
        "forget it",
        "stop",
        "quit",
        "exit",
    ])
    def test_cancel_keywords(self, text):
        assert quick_intent_detect(text) == "cancel"


# ============================================================================
# UX MAGIC INTENTS
# ============================================================================


class TestUndoIntent:
    @pytest.mark.parametrize("text", [
        "undo",
        "go back",
        "wait no",
        "change that",
        "oops",
        "that's wrong",
        "thats wrong",
        "wrong",
        "not right",
    ])
    def test_undo_keywords(self, text):
        assert quick_intent_detect(text) == "undo"


class TestSkipIntent:
    @pytest.mark.parametrize("text", [
        "skip",
        "pass",
        "don't know",
        "dont know",
        "next one",
        "next",
        "don't have",
        "no idea",
    ])
    def test_skip_keywords(self, text):
        assert quick_intent_detect(text) == "skip"

    def test_skip_to_excluded(self):
        """'skip to' is navigation, not skip."""
        result = quick_intent_detect("skip to the next section")
        assert result != "skip"


class TestExplainIntent:
    @pytest.mark.parametrize("text", [
        "why?",
        "why",
        "what's that for?",
        "explain?",
        "what do you mean?",
        "what is?",
        "purpose?",
        "reason?",
    ])
    def test_explain_keywords(self, text):
        assert quick_intent_detect(text) == "explain"


class TestExampleIntent:
    @pytest.mark.parametrize("text", [
        "example?",
        "like what?",
        "show me?",
        "such as?",
        "for instance?",
        "sample?",
        "example",
        "e.g.?",
        "eg?",
    ])
    def test_example_keywords(self, text):
        assert quick_intent_detect(text) == "example"


class TestProgressIntent:
    @pytest.mark.parametrize("text", [
        "how far",
        "how many left",
        "progress",
        "status",
        "how much more",
        "where are we",
    ])
    def test_progress_keywords(self, text):
        assert quick_intent_detect(text) == "progress"


class TestAutofillIntent:
    @pytest.mark.parametrize("text", [
        "same as last time",
        "use my usual",
        "like before",
        "previous",
        "from before",
    ])
    def test_autofill_keywords(self, text):
        assert quick_intent_detect(text) == "autofill"


# ============================================================================
# EDGE CASES
# ============================================================================


class TestEdgeCases:
    def test_empty_string_returns_none(self):
        assert quick_intent_detect("") is None

    def test_single_char_returns_none(self):
        assert quick_intent_detect("a") is None

    def test_whitespace_only_returns_none(self):
        assert quick_intent_detect("   ") is None

    def test_case_insensitivity(self):
        assert quick_intent_detect("SUBMIT") == "submit"
        assert quick_intent_detect("Cancel") == "cancel"
        assert quick_intent_detect("UNDO") == "undo"

    def test_no_match_returns_none(self):
        assert quick_intent_detect("my email is user@example.com") is None

    def test_leading_trailing_whitespace_trimmed(self):
        assert quick_intent_detect("  submit  ") == "submit"


# ============================================================================
# INTENT HELPERS
# ============================================================================


class TestIsLifecycleIntent:
    @pytest.mark.parametrize("intent", ["submit", "stash", "restore", "cancel"])
    def test_lifecycle_intents(self, intent):
        assert is_lifecycle_intent(intent) is True

    @pytest.mark.parametrize("intent", [
        "fill_form", "undo", "skip", "explain", "example", "progress", "autofill", "other",
    ])
    def test_non_lifecycle_intents(self, intent):
        assert is_lifecycle_intent(intent) is False


class TestIsUXIntent:
    @pytest.mark.parametrize("intent", ["undo", "skip", "explain", "example", "progress", "autofill"])
    def test_ux_intents(self, intent):
        assert is_ux_intent(intent) is True

    @pytest.mark.parametrize("intent", [
        "fill_form", "submit", "stash", "restore", "cancel", "other",
    ])
    def test_non_ux_intents(self, intent):
        assert is_ux_intent(intent) is False


class TestHasDataToExtract:
    def test_fill_form_has_data(self):
        assert has_data_to_extract("fill_form") is True

    def test_other_has_data(self):
        assert has_data_to_extract("other") is True

    @pytest.mark.parametrize("intent", [
        "submit", "stash", "restore", "cancel",
        "undo", "skip", "explain", "example", "progress", "autofill",
    ])
    def test_no_data_intents(self, intent):
        assert has_data_to_extract(intent) is False



"""
form_plugin – Guardrails for agent-guided user journeys (Python port).
"""

# Types
from .types import (
    JsonValue,
    UUID,
    FieldStatus,
    SessionStatus,
    FormStatus,
    FieldSource,
    ExternalStatus,
    DependencyCondition,
    FormIntent,
    FormControlOption,
    FormControlFileOptions,
    FormControlDependency,
    FormControlUI,
    FormControlI18n,
    FormControl,
    FormDefinitionUX,
    FormDefinitionTTL,
    FormDefinitionNudge,
    FormDefinitionHooks,
    FormDefinitionI18n,
    FormDefinition,
    FieldFile,
    ExternalFieldState,
    FieldState,
    FieldHistoryEntry,
    SessionEffort,
    FormSession,
    FormSubmission,
    ValidationResult,
    FormContextState,
    ExtractionResult,
    IntentResult,
    FORM_CONTROL_DEFAULTS,
    FORM_DEFINITION_DEFAULTS,
    FORM_SESSION_COMPONENT,
    FORM_SUBMISSION_COMPONENT,
    FORM_AUTOFILL_COMPONENT,
)

# Validation
from .validation import (
    TypeHandler,
    register_type_handler,
    get_type_handler,
    clear_type_handlers,
    validate_field,
    matches_mime_type,
    parse_value,
    format_value,
)

# Intent
from .intent import (
    quick_intent_detect,
    is_lifecycle_intent,
    is_ux_intent,
    has_data_to_extract,
)

# TTL
from .ttl import (
    calculate_ttl,
    should_nudge,
    is_expiring_soon,
    is_expired,
    should_confirm_cancel,
    format_time_remaining,
    format_effort,
)

# Defaults
from .defaults import (
    prettify,
    apply_control_defaults,
    apply_form_defaults,
)

# Builder
from .builder import (
    ControlBuilder,
    FormBuilder,
    Form,
    C,
)

# Template
from .template import (
    TEMPLATE_PATTERN,
    build_template_values,
    render_template,
    resolve_control_templates,
)

# Builtins
from .builtins import (
    BUILTIN_TYPES,
    BUILTIN_TYPE_MAP,
    register_builtin_types,
    get_builtin_type,
    is_builtin_type,
)

__all__ = [
    # Types
    "JsonValue",
    "UUID",
    "FieldStatus",
    "SessionStatus",
    "FormStatus",
    "FieldSource",
    "ExternalStatus",
    "DependencyCondition",
    "FormIntent",
    "FormControlOption",
    "FormControlFileOptions",
    "FormControlDependency",
    "FormControlUI",
    "FormControlI18n",
    "FormControl",
    "FormDefinitionUX",
    "FormDefinitionTTL",
    "FormDefinitionNudge",
    "FormDefinitionHooks",
    "FormDefinitionI18n",
    "FormDefinition",
    "FieldFile",
    "ExternalFieldState",
    "FieldState",
    "FieldHistoryEntry",
    "SessionEffort",
    "FormSession",
    "FormSubmission",
    "ValidationResult",
    "FormContextState",
    "ExtractionResult",
    "IntentResult",
    "FORM_CONTROL_DEFAULTS",
    "FORM_DEFINITION_DEFAULTS",
    "FORM_SESSION_COMPONENT",
    "FORM_SUBMISSION_COMPONENT",
    "FORM_AUTOFILL_COMPONENT",
    # Validation
    "TypeHandler",
    "register_type_handler",
    "get_type_handler",
    "clear_type_handlers",
    "validate_field",
    "matches_mime_type",
    "parse_value",
    "format_value",
    # Intent
    "quick_intent_detect",
    "is_lifecycle_intent",
    "is_ux_intent",
    "has_data_to_extract",
    # TTL
    "calculate_ttl",
    "should_nudge",
    "is_expiring_soon",
    "is_expired",
    "should_confirm_cancel",
    "format_time_remaining",
    "format_effort",
    # Defaults
    "prettify",
    "apply_control_defaults",
    "apply_form_defaults",
    # Builder
    "ControlBuilder",
    "FormBuilder",
    "Form",
    "C",
    # Template
    "TEMPLATE_PATTERN",
    "build_template_values",
    "render_template",
    "resolve_control_templates",
    # Builtins
    "BUILTIN_TYPES",
    "BUILTIN_TYPE_MAP",
    "register_builtin_types",
    "get_builtin_type",
    "is_builtin_type",
]



"""Comprehensive tests for the validation module."""

from __future__ import annotations

import math

import pytest

from elizaos_plugin_form.types import (
    FormControl,
    FormControlFileOptions,
    FormControlOption,
    ValidationResult,
)
from elizaos_plugin_form.validation import (
    TypeHandler,
    clear_type_handlers,
    format_value,
    get_type_handler,
    matches_mime_type,
    parse_value,
    register_type_handler,
    validate_field,
)


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture(autouse=True)
def _clean_handlers():
    """Ensure type handler registry is clean between tests."""
    clear_type_handlers()
    yield
    clear_type_handlers()


def _control(type_: str = "text", **kwargs) -> FormControl:
    """Helper to create a FormControl with sensible defaults."""
    return FormControl(key="test_field", label="Test", type=type_, **kwargs)


# ============================================================================
# REQUIRED vs OPTIONAL
# ============================================================================


class TestRequired:
    def test_required_rejects_none(self):
        result = validate_field(None, _control(required=True))
        assert not result.valid
        assert "required" in (result.error or "").lower()

    def test_required_rejects_empty_string(self):
        result = validate_field("", _control(required=True))
        assert not result.valid

    def test_required_accepts_value(self):
        result = validate_field("hello", _control(required=True))
        assert result.valid

    def test_optional_accepts_none(self):
        result = validate_field(None, _control(required=False))
        assert result.valid

    def test_optional_accepts_empty_string(self):
        result = validate_field("", _control(required=False))
        assert result.valid

    def test_optional_accepts_value(self):
        result = validate_field("hello", _control(required=False))
        assert result.valid


# ============================================================================
# EMAIL VALIDATION
# ============================================================================


class TestEmailValidation:
    def test_valid_email(self):
        assert validate_field("user@example.com", _control("email")).valid

    def test_valid_email_subdomain(self):
        assert validate_field("user@mail.example.co.uk", _control("email")).valid

    def test_invalid_email_no_at(self):
        result = validate_field("userexample.com", _control("email"))
        assert not result.valid
        assert "email" in (result.error or "").lower()

    def test_invalid_email_no_domain(self):
        assert not validate_field("user@", _control("email")).valid

    def test_invalid_email_no_tld(self):
        assert not validate_field("user@example", _control("email")).valid

    def test_invalid_email_spaces(self):
        assert not validate_field("user @example.com", _control("email")).valid

    def test_email_with_plus(self):
        assert validate_field("user+tag@example.com", _control("email")).valid


# ============================================================================
# NUMBER VALIDATION
# ============================================================================


class TestNumberValidation:
    def test_integer(self):
        assert validate_field(42, _control("number")).valid

    def test_float(self):
        assert validate_field(3.14, _control("number")).valid

    def test_string_number(self):
        assert validate_field("42", _control("number")).valid

    def test_nan_rejected(self):
        result = validate_field(float("nan"), _control("number"))
        assert not result.valid

    def test_string_nan_rejected(self):
        result = validate_field("not_a_number", _control("number"))
        assert not result.valid

    def test_min_valid(self):
        assert validate_field(10, _control("number", min=5)).valid

    def test_min_invalid(self):
        result = validate_field(3, _control("number", min=5))
        assert not result.valid
        assert "at least" in (result.error or "").lower()

    def test_max_valid(self):
        assert validate_field(5, _control("number", max=10)).valid

    def test_max_invalid(self):
        result = validate_field(15, _control("number", max=10))
        assert not result.valid
        assert "at most" in (result.error or "").lower()

    def test_comma_formatted(self):
        assert validate_field("1,234", _control("number")).valid

    def test_dollar_formatted(self):
        assert validate_field("$50", _control("number")).valid

    def test_negative_number(self):
        assert validate_field(-5, _control("number")).valid

    def test_zero(self):
        assert validate_field(0, _control("number")).valid


# ============================================================================
# BOOLEAN VALIDATION
# ============================================================================


class TestBooleanValidation:
    @pytest.mark.parametrize("value", [True, False])
    def test_native_bool(self, value):
        assert validate_field(value, _control("boolean")).valid

    @pytest.mark.parametrize("value", ["true", "yes", "1", "on", "True", "YES"])
    def test_truthy_strings(self, value):
        assert validate_field(value, _control("boolean")).valid

    @pytest.mark.parametrize("value", ["false", "no", "0", "off", "False", "NO"])
    def test_falsy_strings(self, value):
        assert validate_field(value, _control("boolean")).valid

    def test_invalid_string(self):
        result = validate_field("maybe", _control("boolean"))
        assert not result.valid
        assert "true or false" in (result.error or "").lower()

    def test_invalid_number(self):
        result = validate_field(42, _control("boolean"))
        assert not result.valid


# ============================================================================
# DATE VALIDATION
# ============================================================================


class TestDateValidation:
    def test_valid_iso_date(self):
        assert validate_field("2024-01-15", _control("date")).valid

    def test_valid_iso_datetime(self):
        assert validate_field("2024-01-15T10:30:00", _control("date")).valid

    def test_invalid_date_string(self):
        result = validate_field("not-a-date", _control("date"))
        assert not result.valid

    def test_min_date(self):
        # Timestamp for 2024-06-01 in ms
        min_ts = 1717200000000
        result = validate_field("2024-01-01", _control("date", min=min_ts))
        assert not result.valid
        assert "too early" in (result.error or "").lower()

    def test_max_date(self):
        # Timestamp for 2024-01-01 in ms
        max_ts = 1704067200000
        result = validate_field("2025-01-01", _control("date", max=max_ts))
        assert not result.valid
        assert "too late" in (result.error or "").lower()

    def test_non_string_non_number(self):
        result = validate_field(["2024-01-01"], _control("date"))
        assert not result.valid


# ============================================================================
# SELECT VALIDATION
# ============================================================================


class TestSelectValidation:
    def _opts(self):
        return [
            FormControlOption(value="us", label="United States"),
            FormControlOption(value="ca", label="Canada"),
        ]

    def test_valid_option(self):
        result = validate_field("us", _control("select", options=self._opts()))
        assert result.valid

    def test_invalid_option(self):
        result = validate_field("mx", _control("select", options=self._opts()))
        assert not result.valid
        assert "available options" in (result.error or "").lower()

    def test_no_options_accepts_anything(self):
        result = validate_field("anything", _control("select"))
        assert result.valid


# ============================================================================
# TEXT VALIDATION
# ============================================================================


class TestTextValidation:
    def test_pattern_match(self):
        assert validate_field("abc123", _control("text", pattern=r"^[a-z0-9]+$")).valid

    def test_pattern_no_match(self):
        result = validate_field("ABC!", _control("text", pattern=r"^[a-z0-9]+$"))
        assert not result.valid
        assert "invalid format" in (result.error or "").lower()

    def test_min_length(self):
        result = validate_field("ab", _control("text", min_length=3))
        assert not result.valid
        assert "at least 3" in (result.error or "")

    def test_max_length(self):
        result = validate_field("toolong", _control("text", max_length=3))
        assert not result.valid
        assert "at most 3" in (result.error or "")

    def test_enum_valid(self):
        assert validate_field("small", _control("text", enum=["small", "medium", "large"])).valid

    def test_enum_invalid(self):
        result = validate_field("xl", _control("text", enum=["small", "medium", "large"]))
        assert not result.valid
        assert "must be one of" in (result.error or "").lower()

    def test_unknown_type_falls_back_to_text(self):
        result = validate_field("hello", _control("custom_unknown"))
        assert result.valid


# ============================================================================
# FILE VALIDATION
# ============================================================================


class TestFileValidation:
    def test_no_file_options_passes(self):
        result = validate_field({"name": "test.txt"}, _control("file"))
        assert result.valid

    def test_max_files_exceeded(self):
        opts = FormControlFileOptions(max_files=2)
        files = [{"name": "a.txt"}, {"name": "b.txt"}, {"name": "c.txt"}]
        result = validate_field(files, _control("file", file=opts))
        assert not result.valid
        assert "Maximum 2" in (result.error or "")

    def test_max_size_exceeded(self):
        opts = FormControlFileOptions(max_size=1024)
        result = validate_field(
            [{"name": "big.txt", "size": 2048, "mimeType": "text/plain"}],
            _control("file", file=opts),
        )
        assert not result.valid
        assert "exceeds" in (result.error or "").lower()

    def test_mime_type_rejected(self):
        opts = FormControlFileOptions(accept=["image/*"])
        result = validate_field(
            [{"name": "doc.pdf", "size": 100, "mimeType": "application/pdf"}],
            _control("file", file=opts),
        )
        assert not result.valid
        assert "not accepted" in (result.error or "").lower()

    def test_mime_type_accepted(self):
        opts = FormControlFileOptions(accept=["image/*"])
        result = validate_field(
            [{"name": "pic.png", "size": 100, "mimeType": "image/png"}],
            _control("file", file=opts),
        )
        assert result.valid


# ============================================================================
# MIME TYPE MATCHING
# ============================================================================


class TestMimeTypeMatching:
    def test_exact_match(self):
        assert matches_mime_type("image/png", "image/png")

    def test_exact_no_match(self):
        assert not matches_mime_type("image/png", "image/jpeg")

    def test_wildcard_match(self):
        assert matches_mime_type("image/png", "image/*")

    def test_wildcard_no_match(self):
        assert not matches_mime_type("application/pdf", "image/*")

    def test_universal_match(self):
        assert matches_mime_type("anything/at-all", "*/*")


# ============================================================================
# TYPE HANDLER REGISTRY
# ============================================================================


class TestTypeHandlerRegistry:
    def test_register_and_get(self):
        handler = TypeHandler(
            validate=lambda v, c: ValidationResult(valid=True),
        )
        register_type_handler("custom", handler)
        assert get_type_handler("custom") is handler

    def test_get_unregistered_returns_none(self):
        assert get_type_handler("nonexistent") is None

    def test_clear(self):
        register_type_handler("temp", TypeHandler())
        clear_type_handlers()
        assert get_type_handler("temp") is None

    def test_custom_handler_priority(self):
        """Custom handler is called before built-in type validation."""
        handler = TypeHandler(
            validate=lambda v, c: ValidationResult(valid=False, error="custom fail"),
        )
        register_type_handler("text", handler)
        result = validate_field("hello", _control("text"))
        assert not result.valid
        assert "custom fail" in (result.error or "")


# ============================================================================
# parse_value
# ============================================================================


class TestParseValue:
    def test_parse_number(self):
        assert parse_value("1,234.56", _control("number")) == 1234.56

    def test_parse_number_dollar(self):
        assert parse_value("$50", _control("number")) == 50.0

    def test_parse_boolean_true(self):
        assert parse_value("yes", _control("boolean")) is True

    def test_parse_boolean_false(self):
        assert parse_value("no", _control("boolean")) is False

    def test_parse_date(self):
        result = parse_value("2024-06-15", _control("date"))
        assert "2024-06-15" in str(result)

    def test_parse_text(self):
        assert parse_value("hello", _control("text")) == "hello"

    def test_parse_email(self):
        assert parse_value("user@example.com", _control("email")) == "user@example.com"

    def test_parse_select(self):
        assert parse_value("option1", _control("select")) == "option1"

    def test_parse_custom_handler(self):
        handler = TypeHandler(parse=lambda v: f"custom:{v}")
        register_type_handler("custom_type", handler)
        result = parse_value("hello", _control("custom_type"))
        assert result == "custom:hello"


# ============================================================================
# format_value
# ============================================================================


class TestFormatValue:
    def test_format_none(self):
        assert format_value(None, _control("text")) == ""

    def test_format_text(self):
        assert format_value("hello", _control("text")) == "hello"

    def test_format_number_int(self):
        result = format_value(1234, _control("number"))
        assert "1,234" in result or "1234" in result

    def test_format_boolean_true(self):
        assert format_value(True, _control("boolean")) == "Yes"

    def test_format_boolean_false(self):
        assert format_value(False, _control("boolean")) == "No"

    def test_format_select_with_label(self):
        opts = [FormControlOption(value="us", label="United States")]
        assert format_value("us", _control("select", options=opts)) == "United States"

    def test_format_select_no_match(self):
        opts = [FormControlOption(value="us", label="United States")]
        assert format_value("ca", _control("select", options=opts)) == "ca"

    def test_format_sensitive_short(self):
        result = format_value("secret", _control("text", sensitive=True))
        assert result == "****"

    def test_format_sensitive_long(self):
        result = format_value("mysecretpassword", _control("text", sensitive=True))
        assert result.startswith("myse")
        assert result.endswith("word")
        assert "..." in result

    def test_format_file_list(self):
        files = [{"name": "a.txt"}, {"name": "b.txt"}]
        result = format_value(files, _control("file"))
        assert "a.txt" in result
        assert "b.txt" in result

    def test_format_file_single(self):
        result = format_value({"name": "doc.pdf"}, _control("file"))
        assert "doc.pdf" in result

    def test_format_custom_handler(self):
        handler = TypeHandler(format=lambda v: f"formatted:{v}")
        register_type_handler("custom_fmt", handler)
        result = format_value("hello", _control("custom_fmt"))
        assert result == "formatted:hello"



"""Comprehensive tests for the TTL module."""

from __future__ import annotations

import time
from unittest.mock import patch

import pytest

from elizaos_plugin_form.types import (
    FormDefinition,
    FormDefinitionNudge,
    FormDefinitionTTL,
    FormSession,
    SessionEffort,
)
from elizaos_plugin_form.ttl import (
    calculate_ttl,
    format_effort,
    format_time_remaining,
    is_expired,
    is_expiring_soon,
    should_confirm_cancel,
    should_nudge,
)


def _now_ms() -> int:
    return int(time.time() * 1000)


def _session(
    time_spent_ms: int = 0,
    last_interaction_ms: int | None = None,
    expires_at: int | None = None,
    nudge_count: int = 0,
    last_nudge_at: int | None = None,
) -> FormSession:
    now = _now_ms()
    return FormSession(
        id="sess-1",
        form_id="form-1",
        effort=SessionEffort(
            interaction_count=1,
            time_spent_ms=time_spent_ms,
            first_interaction_at=now - time_spent_ms,
            last_interaction_at=last_interaction_ms if last_interaction_ms is not None else now,
        ),
        expires_at=expires_at if expires_at is not None else (now + 14 * 24 * 60 * 60 * 1000),
        nudge_count=nudge_count,
        last_nudge_at=last_nudge_at,
    )


def _form(
    min_days: int | None = None,
    max_days: int | None = None,
    effort_multiplier: float | None = None,
    nudge_enabled: bool | None = None,
    nudge_max: int | None = None,
    nudge_after_hours: int | None = None,
) -> FormDefinition:
    ttl = FormDefinitionTTL()
    if min_days is not None:
        ttl.min_days = min_days
    if max_days is not None:
        ttl.max_days = max_days
    if effort_multiplier is not None:
        ttl.effort_multiplier = effort_multiplier
    nudge = FormDefinitionNudge()
    if nudge_enabled is not None:
        nudge.enabled = nudge_enabled
    if nudge_max is not None:
        nudge.max_nudges = nudge_max
    if nudge_after_hours is not None:
        nudge.after_inactive_hours = nudge_after_hours
    return FormDefinition(id="form-1", name="Test Form", ttl=ttl, nudge=nudge)


# ============================================================================
# CALCULATE TTL
# ============================================================================


class TestCalculateTTL:
    def test_no_effort_gets_min_days(self):
        """0 minutes → minDays (14)."""
        session = _session(time_spent_ms=0)
        expiry = calculate_ttl(session)
        now = _now_ms()
        expected_min = now + 14 * 24 * 60 * 60 * 1000
        # Allow 5-second tolerance
        assert abs(expiry - expected_min) < 5000

    def test_moderate_effort(self):
        """30 min * 0.5 = 15 days > 14 min → 15 days."""
        session = _session(time_spent_ms=30 * 60 * 1000)
        expiry = calculate_ttl(session)
        now = _now_ms()
        expected = now + 15 * 24 * 60 * 60 * 1000
        assert abs(expiry - expected) < 5000

    def test_high_effort_capped_at_max(self):
        """4 hours * 0.5 = 120 days, capped at 90."""
        session = _session(time_spent_ms=4 * 60 * 60 * 1000)
        expiry = calculate_ttl(session)
        now = _now_ms()
        expected = now + 90 * 24 * 60 * 60 * 1000
        assert abs(expiry - expected) < 5000

    def test_custom_config(self):
        """Custom config: min=7, max=30, multiplier=1.0."""
        session = _session(time_spent_ms=20 * 60 * 1000)  # 20 min
        form = _form(min_days=7, max_days=30, effort_multiplier=1.0)
        expiry = calculate_ttl(session, form)
        now = _now_ms()
        # 20 min * 1.0 = 20 days
        expected = now + 20 * 24 * 60 * 60 * 1000
        assert abs(expiry - expected) < 5000

    def test_no_form_uses_defaults(self):
        session = _session(time_spent_ms=0)
        expiry = calculate_ttl(session, None)
        now = _now_ms()
        expected = now + 14 * 24 * 60 * 60 * 1000
        assert abs(expiry - expected) < 5000


# ============================================================================
# SHOULD NUDGE
# ============================================================================


class TestShouldNudge:
    def test_nudge_disabled(self):
        session = _session(last_interaction_ms=_now_ms() - 100 * 60 * 60 * 1000)
        form = _form(nudge_enabled=False)
        assert should_nudge(session, form) is False

    def test_max_nudges_reached(self):
        session = _session(
            nudge_count=3,
            last_interaction_ms=_now_ms() - 100 * 60 * 60 * 1000,
        )
        assert should_nudge(session) is False

    def test_not_enough_inactive_time(self):
        """User interacted recently → no nudge."""
        session = _session(last_interaction_ms=_now_ms())
        assert should_nudge(session) is False

    def test_nudge_too_recent(self):
        """Last nudge was <24h ago → no nudge."""
        session = _session(
            last_interaction_ms=_now_ms() - 72 * 60 * 60 * 1000,
            last_nudge_at=_now_ms() - 12 * 60 * 60 * 1000,
        )
        assert should_nudge(session) is False

    def test_should_nudge_all_conditions_met(self):
        """Inactive >48h, no recent nudge, nudge count < max."""
        session = _session(
            last_interaction_ms=_now_ms() - 72 * 60 * 60 * 1000,
            nudge_count=0,
        )
        assert should_nudge(session) is True

    def test_should_nudge_with_previous_nudge_old_enough(self):
        session = _session(
            last_interaction_ms=_now_ms() - 72 * 60 * 60 * 1000,
            nudge_count=1,
            last_nudge_at=_now_ms() - 48 * 60 * 60 * 1000,
        )
        assert should_nudge(session) is True

    def test_custom_after_hours(self):
        session = _session(
            last_interaction_ms=_now_ms() - 12 * 60 * 60 * 1000,
        )
        form = _form(nudge_after_hours=6)
        assert should_nudge(session, form) is True


# ============================================================================
# IS EXPIRING SOON / IS EXPIRED
# ============================================================================


class TestExpiration:
    def test_not_expiring_soon(self):
        session = _session(expires_at=_now_ms() + 48 * 60 * 60 * 1000)
        assert is_expiring_soon(session, 24 * 60 * 60 * 1000) is False

    def test_expiring_soon(self):
        session = _session(expires_at=_now_ms() + 12 * 60 * 60 * 1000)
        assert is_expiring_soon(session, 24 * 60 * 60 * 1000) is True

    def test_not_expired(self):
        session = _session(expires_at=_now_ms() + 1000)
        assert is_expired(session) is False

    def test_expired(self):
        session = _session(expires_at=_now_ms() - 1000)
        assert is_expired(session) is True

    def test_exact_boundary_expired(self):
        """Session that expired exactly now is expired."""
        session = _session(expires_at=_now_ms() - 1)
        assert is_expired(session) is True


# ============================================================================
# SHOULD CONFIRM CANCEL
# ============================================================================


class TestShouldConfirmCancel:
    def test_low_effort_no_confirm(self):
        session = _session(time_spent_ms=2 * 60 * 1000)  # 2 min
        assert should_confirm_cancel(session) is False

    def test_threshold_boundary(self):
        session = _session(time_spent_ms=5 * 60 * 1000)  # exactly 5 min
        assert should_confirm_cancel(session) is False

    def test_high_effort_confirm(self):
        session = _session(time_spent_ms=10 * 60 * 1000)  # 10 min
        assert should_confirm_cancel(session) is True


# ============================================================================
# FORMAT TIME REMAINING
# ============================================================================


class TestFormatTimeRemaining:
    def test_expired(self):
        session = _session(expires_at=_now_ms() - 1000)
        assert format_time_remaining(session) == "expired"

    def test_days_plural(self):
        session = _session(expires_at=_now_ms() + 5 * 24 * 60 * 60 * 1000)
        result = format_time_remaining(session)
        assert "5 days" in result

    def test_day_singular(self):
        session = _session(expires_at=_now_ms() + 1 * 24 * 60 * 60 * 1000 + 1000)
        result = format_time_remaining(session)
        assert "1 day" in result
        assert "days" not in result

    def test_hours_plural(self):
        session = _session(expires_at=_now_ms() + 5 * 60 * 60 * 1000)
        result = format_time_remaining(session)
        assert "5 hours" in result

    def test_hour_singular(self):
        session = _session(expires_at=_now_ms() + 1 * 60 * 60 * 1000 + 1000)
        result = format_time_remaining(session)
        assert "1 hour" in result
        assert "hours" not in result

    def test_minutes_plural(self):
        session = _session(expires_at=_now_ms() + 30 * 60 * 1000)
        result = format_time_remaining(session)
        assert "30 minutes" in result

    def test_minute_singular(self):
        session = _session(expires_at=_now_ms() + 1 * 60 * 1000 + 1000)
        result = format_time_remaining(session)
        assert "1 minute" in result
        assert "minutes" not in result


# ============================================================================
# FORMAT EFFORT
# ============================================================================


class TestFormatEffort:
    def test_just_started(self):
        assert format_effort(_session(time_spent_ms=0)) == "just started"
        assert format_effort(_session(time_spent_ms=30_000)) == "just started"  # 30s

    def test_minutes_singular(self):
        assert format_effort(_session(time_spent_ms=60_000)) == "1 minute"

    def test_minutes_plural(self):
        assert format_effort(_session(time_spent_ms=5 * 60_000)) == "5 minutes"

    def test_hours_exact(self):
        result = format_effort(_session(time_spent_ms=2 * 60 * 60_000))
        assert result == "2 hours"

    def test_hour_singular(self):
        result = format_effort(_session(time_spent_ms=60 * 60_000))
        assert result == "1 hour"

    def test_hours_and_minutes(self):
        result = format_effort(_session(time_spent_ms=90 * 60_000))  # 1h 30m
        assert result == "1h 30m"



"""
Core type definitions for the Form Plugin.

Forms are **guardrails for agent-guided user journeys**.

- FormDefinition = The journey map (what stops are required)
- FormControl = A stop on the journey (what info to collect)
- FormSession = Progress through the journey (where we are)
- FormSubmission = Journey complete (the outcome)
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Literal, Protocol, Union

# ---------------------------------------------------------------------------
# JSON-compatible value type (mirrors JsonValue from @elizaos/core)
# ---------------------------------------------------------------------------
JsonValue = Union[str, int, float, bool, None, list["JsonValue"], dict[str, "JsonValue"]]

# ---------------------------------------------------------------------------
# UUID is just a string alias in the Python port
# ---------------------------------------------------------------------------
UUID = str

# ---------------------------------------------------------------------------
# Field state status
# ---------------------------------------------------------------------------
FieldStatus = Literal["empty", "filled", "uncertain", "invalid", "skipped", "pending"]

# ---------------------------------------------------------------------------
# Session status
# ---------------------------------------------------------------------------
SessionStatus = Literal["active", "ready", "submitted", "stashed", "cancelled", "expired"]

# ---------------------------------------------------------------------------
# Form definition status
# ---------------------------------------------------------------------------
FormStatus = Literal["draft", "active", "deprecated"]

# ---------------------------------------------------------------------------
# Field value source
# ---------------------------------------------------------------------------
FieldSource = Literal["extraction", "autofill", "default", "manual", "correction", "external"]

# ---------------------------------------------------------------------------
# External state status
# ---------------------------------------------------------------------------
ExternalStatus = Literal["pending", "confirmed", "failed", "expired"]

# ---------------------------------------------------------------------------
# Dependency condition
# ---------------------------------------------------------------------------
DependencyCondition = Literal["exists", "equals", "not_equals"]

# ---------------------------------------------------------------------------
# Form intent
# ---------------------------------------------------------------------------
FormIntent = Literal[
    "fill_form",
    "submit",
    "stash",
    "restore",
    "cancel",
    "undo",
    "skip",
    "explain",
    "example",
    "progress",
    "autofill",
    "other",
]


# ============================================================================
# FORM CONTROL – individual field definition
# ============================================================================


@dataclass
class FormControlOption:
    """Select/choice option for select-type fields."""

    value: str
    label: str
    description: str | None = None


@dataclass
class FormControlFileOptions:
    """File upload configuration."""

    accept: list[str] | None = None
    max_size: int | None = None
    max_files: int | None = None


@dataclass
class FormControlDependency:
    """Conditional field dependency."""

    field: str
    condition: DependencyCondition
    value: JsonValue = None


@dataclass
class FormControlUI:
    """UI hints for future frontends."""

    section: str | None = None
    order: int | None = None
    placeholder: str | None = None
    help_text: str | None = None
    widget: str | None = None


@dataclass
class FormControlI18n:
    """Localization for a field."""

    label: str | None = None
    description: str | None = None
    ask_prompt: str | None = None
    help_text: str | None = None


@dataclass
class FormControl:
    """Central field abstraction.

    Each FormControl defines what data to collect, how to validate it,
    how the agent should ask for it, and how to store it.
    """

    # Identity
    key: str
    label: str = ""
    type: str = "text"

    # Behavior
    required: bool = False
    multiple: bool = False
    readonly: bool = False
    hidden: bool = False
    sensitive: bool = False

    # Database binding
    dbbind: str | None = None

    # Validation
    pattern: str | None = None
    min: float | None = None
    max: float | None = None
    min_length: int | None = None
    max_length: int | None = None
    enum: list[str] | None = None

    # Select options
    options: list[FormControlOption] | None = None

    # File options
    file: FormControlFileOptions | None = None

    # Defaults & conditions
    default_value: JsonValue = None
    depends_on: FormControlDependency | None = None

    # Access control
    roles: list[str] | None = None

    # Agent hints
    description: str | None = None
    ask_prompt: str | None = None
    extract_hints: list[str] | None = None
    confirm_threshold: float | None = None
    example: str | None = None

    # UI hints
    ui: FormControlUI | None = None

    # I18n
    i18n: dict[str, FormControlI18n] | None = None

    # Nested fields
    fields: list[FormControl] | None = None

    # Extension
    meta: dict[str, JsonValue] | None = None


# ============================================================================
# FORM DEFINITION – the container for controls
# ============================================================================


@dataclass
class FormDefinitionUX:
    """UX options for the form."""

    allow_undo: bool = True
    allow_skip: bool = True
    max_undo_steps: int = 5
    show_examples: bool = True
    show_explanations: bool = True
    allow_autofill: bool = True


@dataclass
class FormDefinitionTTL:
    """Smart TTL configuration."""

    min_days: int = 14
    max_days: int = 90
    effort_multiplier: float = 0.5


@dataclass
class FormDefinitionNudge:
    """Nudge configuration for stale sessions."""

    enabled: bool = True
    after_inactive_hours: int = 48
    max_nudges: int = 3
    message: str | None = None


@dataclass
class FormDefinitionHooks:
    """Hook configuration (task worker names)."""

    on_start: str | None = None
    on_field_change: str | None = None
    on_ready: str | None = None
    on_submit: str | None = None
    on_cancel: str | None = None
    on_expire: str | None = None


@dataclass
class FormDefinitionI18n:
    """Localization for the form."""

    name: str | None = None
    description: str | None = None


@dataclass
class FormDefinition:
    """The form container – defines a complete form."""

    # Identity
    id: str
    name: str
    controls: list[FormControl] = field(default_factory=list)
    description: str | None = None
    version: int = 1

    # Lifecycle
    status: FormStatus = "active"

    # Permissions
    roles: list[str] | None = None

    # Behavior
    allow_multiple: bool = False

    # UX options
    ux: FormDefinitionUX | None = None

    # TTL (smart retention)
    ttl: FormDefinitionTTL | None = None

    # Nudge
    nudge: FormDefinitionNudge | None = None

    # Hooks
    hooks: FormDefinitionHooks | None = None

    # Debug
    debug: bool = False

    # I18n
    i18n: dict[str, FormDefinitionI18n] | None = None

    # Extension
    meta: dict[str, JsonValue] | None = None


# ============================================================================
# FIELD STATE – runtime state of a single field
# ============================================================================


@dataclass
class FieldFile:
    """File attachment metadata."""

    id: str
    name: str
    mime_type: str
    size: int
    url: str


@dataclass
class ExternalFieldState:
    """State tracking for external/async control types."""

    status: ExternalStatus
    reference: str | None = None
    instructions: str | None = None
    address: str | None = None
    activated_at: int | None = None
    confirmed_at: int | None = None
    external_data: dict[str, JsonValue] | None = None


@dataclass
class FieldState:
    """Runtime state of a field."""

    # Status
    status: FieldStatus = "empty"

    # Value
    value: JsonValue = None

    # Confidence
    confidence: float | None = None
    alternatives: list[JsonValue] | None = None

    # Validation
    error: str | None = None

    # Files
    files: list[FieldFile] | None = None

    # Audit trail
    source: FieldSource | None = None
    message_id: str | None = None
    updated_at: int | None = None
    confirmed_at: int | None = None

    # Composite types
    sub_fields: dict[str, FieldState] | None = None

    # External types
    external_state: ExternalFieldState | None = None

    # Extension
    meta: dict[str, JsonValue] | None = None


# ============================================================================
# FORM SESSION – active form being filled
# ============================================================================


@dataclass
class FieldHistoryEntry:
    """History entry for undo functionality."""

    field: str
    old_value: JsonValue
    new_value: JsonValue
    timestamp: int


@dataclass
class SessionEffort:
    """Effort tracking for smart TTL."""

    interaction_count: int = 0
    time_spent_ms: int = 0
    first_interaction_at: int = 0
    last_interaction_at: int = 0


@dataclass
class FormSession:
    """Active form state."""

    # Identity
    id: str
    form_id: str
    form_version: int | None = None

    # Scoping
    entity_id: UUID = ""
    room_id: UUID = ""

    # Status
    status: SessionStatus = "active"

    # Field data
    fields: dict[str, FieldState] = field(default_factory=dict)

    # History (for undo)
    history: list[FieldHistoryEntry] = field(default_factory=list)

    # Hierarchy
    parent_session_id: str | None = None

    # Context
    context: dict[str, JsonValue] | None = None
    locale: str | None = None

    # Tracking
    last_asked_field: str | None = None
    last_message_id: str | None = None
    cancel_confirmation_asked: bool = False

    # Effort (for smart TTL)
    effort: SessionEffort = field(default_factory=SessionEffort)

    # TTL
    expires_at: int = 0
    expiration_warned: bool = False
    nudge_count: int = 0
    last_nudge_at: int | None = None

    # Timestamps
    created_at: int = 0
    updated_at: int = 0
    submitted_at: int | None = None

    # Extension
    meta: dict[str, JsonValue] | None = None


# ============================================================================
# FORM SUBMISSION – completed form data
# ============================================================================


@dataclass
class FormSubmission:
    """Completed form record."""

    id: str
    form_id: str
    session_id: str
    entity_id: UUID
    values: dict[str, JsonValue] = field(default_factory=dict)
    form_version: int | None = None
    mapped_values: dict[str, JsonValue] | None = None
    files: dict[str, list[FieldFile]] | None = None
    submitted_at: int = 0
    meta: dict[str, JsonValue] | None = None


# ============================================================================
# VALIDATION RESULT
# ============================================================================


@dataclass
class ValidationResult:
    """Standardised validation output."""

    valid: bool
    error: str | None = None


# ============================================================================
# FORM CONTEXT STATE – provider output
# ============================================================================


@dataclass
class FormContextState:
    """Provider output for agent."""

    has_active_form: bool = False
    form_id: str | None = None
    form_name: str | None = None
    progress: int = 0
    filled_fields: list[dict[str, str]] = field(default_factory=list)
    missing_required: list[dict[str, str | None]] = field(default_factory=list)
    uncertain_fields: list[dict[str, JsonValue]] = field(default_factory=list)
    next_field: FormControl | None = None
    status: SessionStatus | None = None
    stashed_count: int = 0
    pending_cancel_confirmation: bool = False
    pending_external_fields: list[dict[str, JsonValue]] = field(default_factory=list)


# ============================================================================
# EXTRACTION / INTENT RESULTS
# ============================================================================


@dataclass
class ExtractionResult:
    """Extraction result for a single field."""

    field: str
    value: JsonValue
    confidence: float
    reasoning: str | None = None
    alternatives: list[JsonValue] | None = None
    is_correction: bool = False


@dataclass
class IntentResult:
    """Combined intent and extraction result."""

    intent: FormIntent
    extractions: list[ExtractionResult] = field(default_factory=list)
    target_form_id: str | None = None


# ============================================================================
# TYPE HANDLER PROTOCOL (legacy)
# ============================================================================


class TypeHandler(Protocol):
    """Custom type behaviour (legacy interface)."""

    def validate(self, value: JsonValue, control: FormControl) -> ValidationResult: ...
    def parse(self, value: str) -> JsonValue: ...
    def format(self, value: JsonValue) -> str: ...
    extraction_prompt: str | None


# ============================================================================
# DEFAULTS
# ============================================================================


FORM_CONTROL_DEFAULTS: dict[str, JsonValue] = {
    "type": "text",
    "required": False,
    "confirm_threshold": 0.8,
}

FORM_DEFINITION_DEFAULTS: dict[str, JsonValue] = {
    "version": 1,
    "status": "active",
    "ux": {
        "allow_undo": True,
        "allow_skip": True,
        "max_undo_steps": 5,
        "show_examples": True,
        "show_explanations": True,
        "allow_autofill": True,
    },
    "ttl": {
        "min_days": 14,
        "max_days": 90,
        "effort_multiplier": 0.5,
    },
    "nudge": {
        "enabled": True,
        "after_inactive_hours": 48,
        "max_nudges": 3,
    },
    "debug": False,
}

FORM_SESSION_COMPONENT = "form_session"
FORM_SUBMISSION_COMPONENT = "form_submission"
FORM_AUTOFILL_COMPONENT = "form_autofill"



"""
Two-tier intent detection for form interactions.

Tier 1 (this module): Fast regex matching for common English phrases.
Tier 2 (LLM): Fallback for non-English, ambiguous, or complex input.
"""

from __future__ import annotations

import re

from .types import FormIntent


def quick_intent_detect(text: str) -> FormIntent | None:
    """Quick intent detection using English keywords.

    Returns ``None`` if no fast-path match found, signalling the caller
    to fall back to LLM-based detection.
    """
    lower = text.lower().strip()

    # Empty or too short
    if len(lower) < 2:
        return None

    # ═══ LIFECYCLE INTENTS ═══

    # Restore
    if re.search(r"\b(resume|continue|pick up where|go back to|get back to)\b", lower):
        return "restore"

    # Submit
    if re.search(
        r"\b(submit|done|finish|send it|that'?s all|i'?m done|complete|all set)\b",
        lower,
    ):
        return "submit"

    # Stash (exclude "save and submit" / "save and send")
    if re.search(
        r"\b(save|stash|later|hold on|pause|save for later|come back|save this)\b",
        lower,
    ):
        if not re.search(r"\b(save and submit|save and send)\b", lower):
            return "stash"

    # Cancel
    if re.search(
        r"\b(cancel|abort|nevermind|never mind|forget it|stop|quit|exit)\b",
        lower,
    ):
        return "cancel"

    # ═══ UX MAGIC INTENTS ═══

    # Undo
    if re.search(
        r"\b(undo|go back|wait no|change that|oops|that'?s wrong|wrong|not right)\b",
        lower,
    ):
        return "undo"

    # Skip (exclude "skip to")
    if re.search(
        r"\b(skip|pass|don'?t know|next one|next|don'?t have|no idea)\b",
        lower,
    ):
        if not re.search(r"\bskip to\b", lower):
            return "skip"

    # Explain
    if re.search(
        r"\b(why|what'?s that for|explain|what do you mean|what is|purpose|reason)\b\??$",
        lower,
        re.IGNORECASE,
    ):
        return "explain"
    if re.match(r"^why\??$", lower, re.IGNORECASE):
        return "explain"

    # Example
    if re.search(
        r"\b(example|like what|show me|such as|for instance|sample)\b\??$",
        lower,
        re.IGNORECASE,
    ):
        return "example"
    if re.match(r"^(example|e\.?g\.?)\??$", lower, re.IGNORECASE):
        return "example"

    # Progress
    if re.search(
        r"\b(how far|how many left|progress|status|how much more|where are we)\b",
        lower,
    ):
        return "progress"

    # Autofill
    if re.search(
        r"\b(same as|last time|use my usual|like before|previous|from before)\b",
        lower,
    ):
        return "autofill"

    return None


# ============================================================================
# INTENT HELPERS
# ============================================================================


def is_lifecycle_intent(intent: FormIntent) -> bool:
    """Check if *intent* is a lifecycle intent (affects session state)."""
    return intent in {"submit", "stash", "restore", "cancel"}


def is_ux_intent(intent: FormIntent) -> bool:
    """Check if *intent* is a UX intent (helper action, no data)."""
    return intent in {"undo", "skip", "explain", "example", "progress", "autofill"}


def has_data_to_extract(intent: FormIntent) -> bool:
    """Check if *intent* likely contains data to extract."""
    return intent in {"fill_form", "other"}



"""Tests for the template module."""

from __future__ import annotations

import pytest

from elizaos_plugin_form.template import (
    TEMPLATE_PATTERN,
    build_template_values,
    render_template,
    resolve_control_templates,
)
from elizaos_plugin_form.types import (
    FieldState,
    FormControl,
    FormControlOption,
    FormSession,
    SessionEffort,
)


def _session(
    fields: dict[str, FieldState] | None = None,
    context: dict[str, object] | None = None,
) -> FormSession:
    return FormSession(
        id="sess-1",
        form_id="form-1",
        fields=fields or {},
        context=context,
        effort=SessionEffort(),
    )


# ============================================================================
# BUILD TEMPLATE VALUES
# ============================================================================


class TestBuildTemplateValues:
    def test_string_value(self):
        session = _session(fields={"name": FieldState(value="Alice")})
        vals = build_template_values(session)
        assert vals["name"] == "Alice"

    def test_number_value(self):
        session = _session(fields={"age": FieldState(value=30)})
        vals = build_template_values(session)
        assert vals["age"] == "30"

    def test_boolean_value(self):
        session = _session(fields={"agree": FieldState(value=True)})
        vals = build_template_values(session)
        assert vals["agree"] == "True"

    def test_none_value_excluded(self):
        session = _session(fields={"empty": FieldState(value=None)})
        vals = build_template_values(session)
        assert "empty" not in vals

    def test_context_values(self):
        session = _session(context={"order_id": "ORD-123"})
        vals = build_template_values(session)
        assert vals["order_id"] == "ORD-123"

    def test_context_number(self):
        session = _session(context={"qty": 5})
        vals = build_template_values(session)
        assert vals["qty"] == "5"

    def test_context_overrides_not_applied(self):
        """Both sources contribute independently."""
        session = _session(
            fields={"email": FieldState(value="a@b.com")},
            context={"ref": "REF-1"},
        )
        vals = build_template_values(session)
        assert vals["email"] == "a@b.com"
        assert vals["ref"] == "REF-1"


# ============================================================================
# RENDER TEMPLATE
# ============================================================================


class TestRenderTemplate:
    def test_simple_substitution(self):
        result = render_template("Hello {{ name }}!", {"name": "Alice"})
        assert result == "Hello Alice!"

    def test_multiple_substitutions(self):
        result = render_template(
            "{{ greeting }}, {{ name }}!",
            {"greeting": "Hi", "name": "Bob"},
        )
        assert result == "Hi, Bob!"

    def test_missing_key_left_in_place(self):
        result = render_template("Hello {{ unknown }}!", {})
        assert result == "Hello {{ unknown }}!"

    def test_none_template(self):
        assert render_template(None, {"key": "val"}) is None

    def test_no_placeholders(self):
        result = render_template("Just text", {"key": "val"})
        assert result == "Just text"

    def test_whitespace_variants(self):
        result = render_template("{{name}} and {{ name }} and {{  name  }}", {"name": "A"})
        assert result == "A and A and A"

    def test_dotted_keys(self):
        result = render_template("{{ user.name }}", {"user.name": "Alice"})
        assert result == "Alice"


# ============================================================================
# RESOLVE CONTROL TEMPLATES
# ============================================================================


class TestResolveControlTemplates:
    def test_label_resolved(self):
        ctrl = FormControl(key="greeting", label="Hello {{ name }}")
        result = resolve_control_templates(ctrl, {"name": "Alice"})
        assert result.label == "Hello Alice"

    def test_description_resolved(self):
        ctrl = FormControl(key="k", label="L", description="For {{ purpose }}")
        result = resolve_control_templates(ctrl, {"purpose": "testing"})
        assert result.description == "For testing"

    def test_ask_prompt_resolved(self):
        ctrl = FormControl(key="k", label="L", ask_prompt="What is your {{ field }}?")
        result = resolve_control_templates(ctrl, {"field": "email"})
        assert result.ask_prompt == "What is your email?"

    def test_example_resolved(self):
        ctrl = FormControl(key="k", label="L", example="{{ sample }}")
        result = resolve_control_templates(ctrl, {"sample": "test@example.com"})
        assert result.example == "test@example.com"

    def test_extract_hints_resolved(self):
        ctrl = FormControl(key="k", label="L", extract_hints=["{{ hint1 }}", "static"])
        result = resolve_control_templates(ctrl, {"hint1": "dynamic"})
        assert result.extract_hints == ["dynamic", "static"]

    def test_options_resolved(self):
        ctrl = FormControl(
            key="k",
            label="L",
            options=[
                FormControlOption(value="v1", label="{{ opt_label }}", description="{{ opt_desc }}"),
            ],
        )
        result = resolve_control_templates(ctrl, {"opt_label": "Option 1", "opt_desc": "Desc 1"})
        assert result.options is not None
        assert result.options[0].label == "Option 1"
        assert result.options[0].description == "Desc 1"

    def test_nested_fields_resolved(self):
        child = FormControl(key="child", label="{{ child_label }}")
        parent = FormControl(key="parent", label="Parent", fields=[child])
        result = resolve_control_templates(parent, {"child_label": "Child Label"})
        assert result.fields is not None
        assert result.fields[0].label == "Child Label"

    def test_original_preserved_on_no_match(self):
        ctrl = FormControl(key="k", label="No templates here")
        result = resolve_control_templates(ctrl, {"unused": "val"})
        assert result.label == "No templates here"

    def test_key_and_type_unchanged(self):
        ctrl = FormControl(key="mykey", label="L", type="email")
        result = resolve_control_templates(ctrl, {"mykey": "ignored"})
        assert result.key == "mykey"
        assert result.type == "email"



"""
Fluent builder API for defining forms and controls.

Shorthand exports:
- ``Form`` = ``FormBuilder``
- ``C`` = ``ControlBuilder``

Example::

    form = Form.create('contact') \\
        .name('Contact Form') \\
        .control(C.email('email').required()) \\
        .control(C.text('message').required()) \\
        .build()
"""

from __future__ import annotations

import re
from typing import Union

from .types import (
    FormControl,
    FormControlDependency,
    FormControlFileOptions,
    FormControlI18n,
    FormControlOption,
    FormControlUI,
    FormDefinition,
    FormDefinitionHooks,
    FormDefinitionNudge,
    FormDefinitionTTL,
    FormDefinitionUX,
    JsonValue,
)


def _prettify(key: str) -> str:
    spaced = re.sub(r"[-_]", " ", key)
    return re.sub(r"\b\w", lambda m: m.group().upper(), spaced)


# ============================================================================
# CONTROL BUILDER
# ============================================================================


class ControlBuilder:
    """Fluent builder for :class:`FormControl`."""

    def __init__(self, key: str) -> None:
        self._key = key
        self._type: str | None = None
        self._label: str | None = None
        self._required: bool | None = None
        self._multiple: bool | None = None
        self._readonly: bool | None = None
        self._hidden: bool | None = None
        self._sensitive: bool | None = None
        self._dbbind: str | None = None
        self._pattern: str | None = None
        self._min: float | None = None
        self._max: float | None = None
        self._min_length: int | None = None
        self._max_length: int | None = None
        self._enum: list[str] | None = None
        self._options: list[FormControlOption] | None = None
        self._file: FormControlFileOptions | None = None
        self._default_value: JsonValue = None
        self._depends_on: FormControlDependency | None = None
        self._roles: list[str] | None = None
        self._description: str | None = None
        self._ask_prompt: str | None = None
        self._extract_hints: list[str] | None = None
        self._confirm_threshold: float | None = None
        self._example: str | None = None
        self._ui: FormControlUI | None = None
        self._i18n: dict[str, FormControlI18n] | None = None
        self._meta: dict[str, JsonValue] | None = None
        self._fields: list[FormControl] | None = None

    # ═══ STATIC FACTORIES ═══

    @staticmethod
    def field(key: str) -> ControlBuilder:
        return ControlBuilder(key)

    @staticmethod
    def text(key: str) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "text"
        return b

    @staticmethod
    def email(key: str) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "email"
        return b

    @staticmethod
    def number(key: str) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "number"
        return b

    @staticmethod
    def boolean_(key: str) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "boolean"
        return b

    @staticmethod
    def select(key: str, options: list[FormControlOption] | None = None) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "select"
        if options:
            b._options = options
        return b

    @staticmethod
    def date(key: str) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "date"
        return b

    @staticmethod
    def file(key: str) -> ControlBuilder:
        b = ControlBuilder(key)
        b._type = "file"
        return b

    # ═══ TYPE ═══

    def type(self, type_name: str) -> ControlBuilder:
        self._type = type_name
        return self

    # ═══ BEHAVIOR ═══

    def required(self) -> ControlBuilder:
        self._required = True
        return self

    def optional(self) -> ControlBuilder:
        self._required = False
        return self

    def hidden(self) -> ControlBuilder:
        self._hidden = True
        return self

    def sensitive(self) -> ControlBuilder:
        self._sensitive = True
        return self

    def readonly(self) -> ControlBuilder:
        self._readonly = True
        return self

    def multiple(self) -> ControlBuilder:
        self._multiple = True
        return self

    # ═══ VALIDATION ═══

    def pattern(self, regex: str) -> ControlBuilder:
        self._pattern = regex
        return self

    def min(self, n: float) -> ControlBuilder:
        self._min = n
        return self

    def max(self, n: float) -> ControlBuilder:
        self._max = n
        return self

    def min_length(self, n: int) -> ControlBuilder:
        self._min_length = n
        return self

    def max_length(self, n: int) -> ControlBuilder:
        self._max_length = n
        return self

    def enum(self, values: list[str]) -> ControlBuilder:
        self._enum = values
        return self

    # ═══ AGENT HINTS ═══

    def label(self, lbl: str) -> ControlBuilder:
        self._label = lbl
        return self

    def ask(self, prompt: str) -> ControlBuilder:
        self._ask_prompt = prompt
        return self

    def description(self, desc: str) -> ControlBuilder:
        self._description = desc
        return self

    def hint(self, *hints: str) -> ControlBuilder:
        self._extract_hints = list(hints)
        return self

    def example(self, value: str) -> ControlBuilder:
        self._example = value
        return self

    def confirm_threshold(self, n: float) -> ControlBuilder:
        self._confirm_threshold = n
        return self

    # ═══ FILE OPTIONS ═══

    def accept(self, mime_types: list[str]) -> ControlBuilder:
        if self._file is None:
            self._file = FormControlFileOptions()
        self._file.accept = mime_types
        return self

    def max_size(self, num_bytes: int) -> ControlBuilder:
        if self._file is None:
            self._file = FormControlFileOptions()
        self._file.max_size = num_bytes
        return self

    def max_files(self, n: int) -> ControlBuilder:
        if self._file is None:
            self._file = FormControlFileOptions()
        self._file.max_files = n
        return self

    # ═══ ACCESS ═══

    def roles(self, *role_names: str) -> ControlBuilder:
        self._roles = list(role_names)
        return self

    def default(self, value: JsonValue) -> ControlBuilder:
        self._default_value = value
        return self

    def depends_on(
        self,
        field_key: str,
        condition: str = "exists",
        value: JsonValue = None,
    ) -> ControlBuilder:
        self._depends_on = FormControlDependency(field=field_key, condition=condition, value=value)  # type: ignore[arg-type]
        return self

    def dbbind(self, column_name: str) -> ControlBuilder:
        self._dbbind = column_name
        return self

    # ═══ UI ═══

    def section(self, name: str) -> ControlBuilder:
        if self._ui is None:
            self._ui = FormControlUI()
        self._ui.section = name
        return self

    def order(self, n: int) -> ControlBuilder:
        if self._ui is None:
            self._ui = FormControlUI()
        self._ui.order = n
        return self

    def placeholder(self, text: str) -> ControlBuilder:
        if self._ui is None:
            self._ui = FormControlUI()
        self._ui.placeholder = text
        return self

    def help_text(self, text: str) -> ControlBuilder:
        if self._ui is None:
            self._ui = FormControlUI()
        self._ui.help_text = text
        return self

    def widget(self, widget_type: str) -> ControlBuilder:
        if self._ui is None:
            self._ui = FormControlUI()
        self._ui.widget = widget_type
        return self

    # ═══ I18N ═══

    def i18n(
        self,
        locale: str,
        translations: dict[str, str | None],
    ) -> ControlBuilder:
        if self._i18n is None:
            self._i18n = {}
        self._i18n[locale] = FormControlI18n(
            label=translations.get("label"),
            description=translations.get("description"),
            ask_prompt=translations.get("ask_prompt"),
            help_text=translations.get("help_text"),
        )
        return self

    # ═══ META ═══

    def meta(self, key: str, value: JsonValue) -> ControlBuilder:
        if self._meta is None:
            self._meta = {}
        self._meta[key] = value
        return self

    # ═══ BUILD ═══

    def build(self) -> FormControl:
        """Build the final ``FormControl``."""
        return FormControl(
            key=self._key,
            label=self._label or _prettify(self._key),
            type=self._type or "text",
            required=self._required or False,
            multiple=self._multiple or False,
            readonly=self._readonly or False,
            hidden=self._hidden or False,
            sensitive=self._sensitive or False,
            dbbind=self._dbbind,
            pattern=self._pattern,
            min=self._min,
            max=self._max,
            min_length=self._min_length,
            max_length=self._max_length,
            enum=self._enum,
            options=self._options,
            file=self._file,
            default_value=self._default_value,
            depends_on=self._depends_on,
            roles=self._roles,
            description=self._description,
            ask_prompt=self._ask_prompt,
            extract_hints=self._extract_hints,
            confirm_threshold=self._confirm_threshold,
            example=self._example,
            ui=self._ui,
            i18n=self._i18n,
            meta=self._meta,
            fields=self._fields,
        )


# ============================================================================
# FORM BUILDER
# ============================================================================


class FormBuilder:
    """Fluent builder for :class:`FormDefinition`."""

    def __init__(self, form_id: str) -> None:
        self._id = form_id
        self._name: str | None = None
        self._description: str | None = None
        self._version: int | None = None
        self._controls: list[FormControl] = []
        self._roles: list[str] | None = None
        self._allow_multiple: bool = False
        self._ux: FormDefinitionUX | None = None
        self._ttl_cfg: FormDefinitionTTL | None = None
        self._nudge_cfg: FormDefinitionNudge | None = None
        self._hooks: FormDefinitionHooks | None = None
        self._debug: bool = False
        self._i18n: dict[str, dict[str, str | None]] | None = None
        self._meta: dict[str, JsonValue] | None = None

    # ═══ STATIC FACTORY ═══

    @staticmethod
    def create(form_id: str) -> FormBuilder:
        return FormBuilder(form_id)

    # ═══ METADATA ═══

    def name(self, name: str) -> FormBuilder:
        self._name = name
        return self

    def description(self, desc: str) -> FormBuilder:
        self._description = desc
        return self

    def version(self, v: int) -> FormBuilder:
        self._version = v
        return self

    # ═══ CONTROLS ═══

    def control(self, builder: Union[ControlBuilder, FormControl]) -> FormBuilder:
        ctrl = builder.build() if isinstance(builder, ControlBuilder) else builder
        self._controls.append(ctrl)
        return self

    def controls(self, *builders: Union[ControlBuilder, FormControl]) -> FormBuilder:
        for b in builders:
            self.control(b)
        return self

    def required(self, *keys: str) -> FormBuilder:
        for key in keys:
            self.control(ControlBuilder.field(key).required())
        return self

    def optional(self, *keys: str) -> FormBuilder:
        for key in keys:
            self.control(ControlBuilder.field(key))
        return self

    # ═══ PERMISSIONS ═══

    def roles(self, *role_names: str) -> FormBuilder:
        self._roles = list(role_names)
        return self

    def allow_multiple(self) -> FormBuilder:
        self._allow_multiple = True
        return self

    # ═══ UX ═══

    def no_undo(self) -> FormBuilder:
        if self._ux is None:
            self._ux = FormDefinitionUX()
        self._ux.allow_undo = False
        return self

    def no_skip(self) -> FormBuilder:
        if self._ux is None:
            self._ux = FormDefinitionUX()
        self._ux.allow_skip = False
        return self

    def no_autofill(self) -> FormBuilder:
        if self._ux is None:
            self._ux = FormDefinitionUX()
        self._ux.allow_autofill = False
        return self

    def max_undo_steps(self, n: int) -> FormBuilder:
        if self._ux is None:
            self._ux = FormDefinitionUX()
        self._ux.max_undo_steps = n
        return self

    # ═══ TTL ═══

    def ttl(
        self,
        min_days: int | None = None,
        max_days: int | None = None,
        effort_multiplier: float | None = None,
    ) -> FormBuilder:
        if self._ttl_cfg is None:
            self._ttl_cfg = FormDefinitionTTL()
        if min_days is not None:
            self._ttl_cfg.min_days = min_days
        if max_days is not None:
            self._ttl_cfg.max_days = max_days
        if effort_multiplier is not None:
            self._ttl_cfg.effort_multiplier = effort_multiplier
        return self

    # ═══ NUDGE ═══

    def no_nudge(self) -> FormBuilder:
        if self._nudge_cfg is None:
            self._nudge_cfg = FormDefinitionNudge()
        self._nudge_cfg.enabled = False
        return self

    def nudge_after(self, hours: int) -> FormBuilder:
        if self._nudge_cfg is None:
            self._nudge_cfg = FormDefinitionNudge()
        self._nudge_cfg.after_inactive_hours = hours
        return self

    def nudge_message(self, message: str) -> FormBuilder:
        if self._nudge_cfg is None:
            self._nudge_cfg = FormDefinitionNudge()
        self._nudge_cfg.message = message
        return self

    # ═══ HOOKS ═══

    def on_start(self, worker_name: str) -> FormBuilder:
        if self._hooks is None:
            self._hooks = FormDefinitionHooks()
        self._hooks.on_start = worker_name
        return self

    def on_field_change(self, worker_name: str) -> FormBuilder:
        if self._hooks is None:
            self._hooks = FormDefinitionHooks()
        self._hooks.on_field_change = worker_name
        return self

    def on_ready(self, worker_name: str) -> FormBuilder:
        if self._hooks is None:
            self._hooks = FormDefinitionHooks()
        self._hooks.on_ready = worker_name
        return self

    def on_submit(self, worker_name: str) -> FormBuilder:
        if self._hooks is None:
            self._hooks = FormDefinitionHooks()
        self._hooks.on_submit = worker_name
        return self

    def on_cancel(self, worker_name: str) -> FormBuilder:
        if self._hooks is None:
            self._hooks = FormDefinitionHooks()
        self._hooks.on_cancel = worker_name
        return self

    def on_expire(self, worker_name: str) -> FormBuilder:
        if self._hooks is None:
            self._hooks = FormDefinitionHooks()
        self._hooks.on_expire = worker_name
        return self

    # ═══ DEBUG ═══

    def debug(self) -> FormBuilder:
        self._debug = True
        return self

    # ═══ I18N ═══

    def i18n(self, locale: str, translations: dict[str, str | None]) -> FormBuilder:
        if self._i18n is None:
            self._i18n = {}
        self._i18n[locale] = translations
        return self

    # ═══ META ═══

    def meta(self, key: str, value: JsonValue) -> FormBuilder:
        if self._meta is None:
            self._meta = {}
        self._meta[key] = value
        return self

    # ═══ BUILD ═══

    def build(self) -> FormDefinition:
        """Build the final ``FormDefinition``."""
        from .types import FormDefinitionI18n as _FI18n

        i18n_map: dict[str, _FI18n] | None = None
        if self._i18n:
            i18n_map = {
                loc: _FI18n(name=vals.get("name"), description=vals.get("description"))
                for loc, vals in self._i18n.items()
            }

        return FormDefinition(
            id=self._id,
            name=self._name or _prettify(self._id),
            controls=self._controls,
            description=self._description,
            version=self._version or 1,
            status="active",
            roles=self._roles,
            allow_multiple=self._allow_multiple,
            ux=self._ux,
            ttl=self._ttl_cfg,
            nudge=self._nudge_cfg,
            hooks=self._hooks,
            debug=self._debug,
            i18n=i18n_map,
            meta=self._meta,
        )


# ============================================================================
# SHORTHAND EXPORTS
# ============================================================================

Form = FormBuilder
"""Alias for :class:`FormBuilder`."""

C = ControlBuilder
"""Alias for :class:`ControlBuilder`."""
